// Copyright (c) 2015, Electric Cloud, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   - Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
//   - Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// (As an historical note, the above license instantiates
// the "BSD 2-Clause License" template that was presented
// by the Open Source Initiative at
//
//   http://www.opensource.org/licenses/bsd-license.php
//
// on 2012-02-23.)

// Summary:
//
// This header defines C++ wrappers for interfaces in:
//
//     "../interface/vrpsx_allocator.h"
//
// NOTE: This convenience header does not affect the
// definition of any any binary interface; see README.txt.

#ifndef VRPSX_ALLOCATOR_HPP
#define VRPSX_ALLOCATOR_HPP 1

#include "../interface/vrpsx_allocator.h"
#include "vrpsx_cpp_wrapper.hpp"
#include "vrpsx_move.hpp"


// We do not know where "move"


namespace vrpsx {
    // We use this kind of smart pointer for allocated memory.

    using boost::interprocess::unique_ptr;

    VRPSX_CPP_WRAPPER(allocator_t);

    // unique_ptr< M, free_memory_t > is the correct type for
    // smart pointers to memory of type M allocated by an allocator_t.
    //
    // NOTE: Due to the way unique_ptr works, at least in Boost 1.42.0,
    // we must allow default construction of free_memory_t.  But merely
    // invoking a default-constructed free_memory_t, even on NULL, will
    // cause your program to crash.
    //
    // Nevertheless, it is safe to default-construct an instance
    // of unique_ptr< M, free_memory_t > because unique_ptr does
    // not actually invoke the deleter on a NULL pointer value.

    class free_memory_t
    {
    public:
        free_memory_t(
            allocator_t *allocator = 0)
            : mAllocator(from_cpp(allocator))
        {}

        void operator()(
            void *memory)
        {
            mAllocator->vtbl->free_memory(mAllocator, memory);
        }

    private:
        vrpsx_allocator_t *mAllocator;
    };

    class allocator_t
    {
    public:
        unique_ptr< void, free_memory_t > allocate_memory(
            size_t size)
        {
            vrpsx_allocator_t *iface = from_cpp(this);
            return unique_ptr< void, free_memory_t >(
                iface->vtbl->allocate_memory(iface, size),
                free_memory_t(this));
        }

    private:
        char mEnsureNonzeroSize;  // helps when inheriting from
                                  // this class at a nonzero offset
    };
}


#endif /* VRPSX_ALLOCATOR_HPP */
