// Copyright (c) 2015, Electric Cloud, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   - Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
//   - Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// (As an historical note, the above license instantiates
// the "BSD 2-Clause License" template that was presented
// by the Open Source Initiative at
//
//   http://www.opensource.org/licenses/bsd-license.php
//
// on 2012-02-23.)

// Summary:
//
// This header defines C++ wrappers for interfaces in:
//
//     "../interface/vrpsx_system.h"
//
// NOTE: This convenience header does not affect the
// definition of any any binary interface; see README.txt.
//
// In general, methods that end in "_nt" are "nothrow" variants
// that return error numbers instead of throwing exceptions.

#ifndef VRPSX_SYSTEM_HPP
#define VRPSX_SYSTEM_HPP 1

#include "../interface/vrpsx_system.h"
#include "vrpsx_allocator.hpp"
#include "vrpsx_system_error.hpp"
#include "vrpsx_single_owner.hpp"
#include <algorithm>
#include <stdexcept>


namespace vrpsx {
    VRPSX_CPP_WRAPPER(system_t);

    inline intrusive_ptr< system_error_t > error_acquire(
        system_t& system,
        vrpsx_errno_t ern)
    {
        vrpsx_system_t *iface = from_cpp(&system);
        vrpsx_system_error_t *out_error;
        iface->vtbl->error_acquire(iface, ern, &out_error);
        return intrusive_ptr< system_error_t >(to_cpp(out_error), false);
    }

    // Templated merely so that we can define it in a header.
    // The linker should eliminate redundant definitions.

    template< typename unused >
    std::string prefixed_error_impl(
        const char *prefix,
        system_t& system,
        vrpsx_errno_t ern)
    {
        // Get an error object and cache a pointer to its text,
        // which remains valid until we release our reference.

        intrusive_ptr< system_error_t > error(error_acquire(system, ern));
        const char *errorText = error->text();

        // Allocate exactly the required size of string.

        std::size_t errorLength = std::char_traits< char >::length(errorText);
        std::size_t prefixLength = std::char_traits< char >::length(prefix);
        std::string result(prefixLength + 2 + errorLength, '\0');
        std::string::iterator it = result.begin();

        // Copy the prefix.

        it = std::copy(prefix, prefix + prefixLength, it);

        // Copy the ": ".

        *it++ = ':';
        *it++ = ' ';

        // Copy the error text.

        std::copy(errorText, errorText + errorLength, it);

        return result;
    }

    // Equivalent to:
    //     std::string(prefix) + ": " + error_acquire(system, ern)->text()

    inline std::string prefixed_error(
        const char *prefix,
        system_t& system,
        vrpsx_errno_t ern)
    {
        return prefixed_error_impl< void >(prefix, system, ern);
    }

    class system_exception_t : public std::runtime_error
    {
    public:
        system_exception_t(
            const char *prefix,
            system_t& system,
            vrpsx_errno_t ern)
            : std::runtime_error(prefixed_error(prefix, system, ern)),
              mNumber(ern)
        {}

        vrpsx_errno_t number() const
        {
            return mNumber;
        }

    private:
        vrpsx_errno_t mNumber;
    };

    // Templated merely so that we can define it in a header.
    // The linker should eliminate redundant definitions.

    template< typename unused >
    void unconditional_throw_impl(
        const char *prefix,
        system_t& system,
        vrpsx_errno_t ern)
    {
        throw system_exception_t(prefix, system, ern);
    }

    inline void unconditional_throw(
        const char *prefix,
        system_t& system,
        vrpsx_errno_t ern)
    {
        unconditional_throw_impl< void >(prefix, system, ern);
    }

    // Templated merely so that we can define it in a header.
    // The linker should eliminate redundant definitions.

    template< typename unused >
    void disown_recover_impl(
        const char *prefix,
        vrpsx_system_t *raw_sroe,  // see "System Reference On Error"
        vrpsx_errno_t ern)
    {
        intrusive_ptr< system_t > sroe(to_cpp(raw_sroe), false);
        if (! std::uncaught_exception()) {
            unconditional_throw(prefix, *sroe, ern);
        }
    }

    inline void disown_recover(
        const char *prefix,
        vrpsx_system_t *raw_sroe,  // see "System Reference On Error"
        vrpsx_errno_t ern)
    {
        disown_recover_impl< void >(prefix, raw_sroe, ern);
    }

    inline void disown_check(
        const char *prefix,
        vrpsx_system_t *raw_sroe,  // see "System Reference On Error"
        vrpsx_errno_t ern)
    {
        if (ern) {
            disown_recover(prefix, raw_sroe, ern);
        }
    }


    VRPSX_CPP_WRAPPER(system_associate_t);

    class system_associate_t
    {
    private:
        template< typename unused >
        void unconditional_throw_impl(
            const char *prefix,
            vrpsx_errno_t ern) const;

    public:
        void throw_on_error(
            const char *prefix,
            vrpsx_errno_t ern) const
        {
            if (ern) {
                unconditional_throw(prefix, ern);
            }
        }

        void unconditional_throw(
            const char *prefix,
            vrpsx_errno_t ern) const
        {
            unconditional_throw_impl< void >(prefix, ern);
        }

        system_t *get_system() const
        {
            vrpsx_system_associate_t const *iface = from_cpp(this);
            return to_cpp(iface->vtbl->get_system(iface));
        }
    };

    // Templated merely so that we can define it in a header.
    // The linker should eliminate redundant definitions.

    template< typename unused >
    void system_associate_t::unconditional_throw_impl(
        const char *prefix,
        vrpsx_errno_t ern) const
    {
        vrpsx::unconditional_throw(prefix, *get_system(), ern);
    }


    VRPSX_CPP_WRAPPER(utsname_t);

    class utsname_t
    {
    public:
        void disown_mt()
        {
            vrpsx_utsname_t *iface = from_cpp(this);
            iface->vtbl->disown(&iface);
        }

        char const *machine() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->machine(iface);
        }

        char const *nodename() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->nodename(iface);
        }

        char const *release() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->release(iface);
        }

        char const *sysname() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->sysname(iface);
        }

        char const *version() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->version(iface);
        }

        char const *virtualization() const
        {
            vrpsx_utsname_t const *iface = from_cpp(this);
            return iface->vtbl->virtualization(iface);
        }
    };


    VRPSX_CPP_WRAPPER(passwd_t);

    class passwd_t
    {
    public:
        void disown_mt()
        {
            vrpsx_passwd_t *iface = from_cpp(this);
            iface->vtbl->disown(&iface);
        }

        char const *dir() const
        {
            vrpsx_passwd_t const *iface = from_cpp(this);
            return iface->vtbl->dir(iface);
        }

        char const *name() const
        {
            vrpsx_passwd_t const *iface = from_cpp(this);
            return iface->vtbl->name(iface);
        }

        char const *shell() const
        {
            vrpsx_passwd_t const *iface = from_cpp(this);
            return iface->vtbl->shell(iface);
        }

        vrpsx_gid_t gid() const
        {
            vrpsx_passwd_t const *iface = from_cpp(this);
            return iface->vtbl->gid(iface);
        }

        vrpsx_uid_t uid() const
        {
            vrpsx_passwd_t const *iface = from_cpp(this);
            return iface->vtbl->uid(iface);
        }
    };


    // NOTE: If you need async-signal-safe signal set
    // manipulation, then use vrpsx_sigset_t directly.

    VRPSX_CPP_WRAPPER(sigset_t);

    class sigset_t : public system_associate_t
    {
    public:
        /* Assigns the signal set of *source to that of this iface.
         * State associated with extension ifaces is not copied.
         * The two ifaces must have originated in the same vrpsx_system_t.
         */
        void assign(
            sigset_t const& source)
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            iface->vtbl->assign_sigset(iface, from_cpp(&source));
        }

        void swap(
            sigset_t& other)
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            iface->vtbl->swap_sigset(iface, from_cpp(&other));
        }

        vrpsx_errno_t empty_nt()
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            return iface->vtbl->empty(iface);
        }

        void empty()
        {
            throw_on_error(
                "vrpsx::sigset::empty",
                empty_nt());
        }

        vrpsx_errno_t fill_nt()
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            return iface->vtbl->fill(iface);
        }

        void fill()
        {
            throw_on_error(
                "vrpsx::sigset::fill",
                fill_nt());
        }

        vrpsx_errno_t del_nt(
            vrpsx_signo_t signo)
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            return iface->vtbl->del(iface, signo);
        }

        void del(
            vrpsx_signo_t signo)
        {
            throw_on_error(
                "vrpsx::sigset::del",
                del_nt(signo));
        }

        vrpsx_errno_t add_nt(
            vrpsx_signo_t signo)
        {
            vrpsx_sigset_t *iface = from_cpp(this);
            return iface->vtbl->add(iface, signo);
        }

        void add(
            vrpsx_signo_t signo)
        {
            throw_on_error(
                "vrpsx::sigset::add",
                add_nt(signo));
        }

        vrpsx_errno_t ismember_nt(
            vrpsx_signo_t signo,
            bool& out_is_member) const
        {
            vrpsx_sigset_t const *iface = from_cpp(this);
            int raw_out_is_member;
            vrpsx_errno_t ern = iface->vtbl->ismember(
                iface, signo, &raw_out_is_member);
            out_is_member = (raw_out_is_member != 0);
            return ern;
        }

        bool ismember(
            vrpsx_signo_t signo) const
        {
            bool result;
            throw_on_error(
                "vrpsx::sigset::ismember",
                ismember_nt(signo, result));
            return result;
        }
    };


    // NOTE: If you need async-signal-safe signal set
    // manipulation, then use vrpsx_sigaction_t directly.

    VRPSX_CPP_WRAPPER(sigaction_t);

    class sigaction_t : public system_associate_t
    {
    public:
        /* Assigns the signal set of *source to that of this iface.
         * State associated with extension ifaces is not copied.
         * The two ifaces must have originated in the same vrpsx_system_t.
         */
        void assign(
            sigaction_t const& source)
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            iface->vtbl->assign_sigaction(iface, from_cpp(&source));
        }

        void swap(
            sigaction_t& other)
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            iface->vtbl->swap_sigaction(iface, from_cpp(&other));
        }

        vrpsx_sa_handler_t *get_handler() const
        {
            vrpsx_sigaction_t const *iface = from_cpp(this);
            return iface->vtbl->get_handler(iface);
        }

        void set_handler(
            vrpsx_sa_handler_t *handler)
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            iface->vtbl->set_handler(iface, handler);
        }

        vrpsx_sa_sigaction_t *get_sigaction() const
        {
            vrpsx_sigaction_t const *iface = from_cpp(this);
            return iface->vtbl->get_sigaction(iface);
        }

        void set_sigaction(
            vrpsx_sa_sigaction_t *handler)
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            iface->vtbl->set_sigaction(iface, handler);
        }

        sigset_t const *get_mask() const
        {
            vrpsx_sigaction_t const *iface = from_cpp(this);
            return to_cpp(iface->vtbl->get_mask_const(iface));
        }

        sigset_t *get_mask()
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            return to_cpp(iface->vtbl->get_mask(iface));
        }

        vrpsx_sa_flags_t get_flags() const
        {
            vrpsx_sigaction_t const *iface = from_cpp(this);
            return iface->vtbl->get_flags(iface);
        }

        vrpsx_errno_t set_flags_nt(
            vrpsx_sa_flags_t flags)
        {
            vrpsx_sigaction_t *iface = from_cpp(this);
            return iface->vtbl->set_flags(iface, flags);
        }

        void set_flags(
            vrpsx_sa_flags_t flags)
        {
            throw_on_error(
                "vrpsx::sigaction::set_flags",
                set_flags_nt(flags));
        }
    };


    VRPSX_CPP_WRAPPER(thread_attr_t);

    class thread_attr_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_thread_attr_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::thread_attr::disown",
                sroe, ern);
        }

        vrpsx_errno_t create_nt(
            vrpsx_thread_t& thread,
            vrpsx_thread_function_t *function,
            void *argument) const
        {
            vrpsx_thread_attr_t const *iface = from_cpp(this);
            return iface->vtbl->create(
                iface, &thread, function, argument);
        }

        vrpsx_thread_t create(
            vrpsx_thread_function_t *function,
            void *argument) const
        {
            vrpsx_thread_t result;
            throw_on_error(
                "vrpsx::thread_attr::create",
                create_nt(result, function, argument));
            return result;
        }

        vrpsx_errno_t getdetachstate_nt(
            int& detach_state) const
        {
            vrpsx_thread_attr_t const *iface = from_cpp(this);
            return iface->vtbl->getdetachstate(iface, &detach_state);
        }

        int getdetachstate() const
        {
            int result;
            throw_on_error(
                "vrpsx::thread_attr::getdetachstate",
                getdetachstate_nt(result));
            return result;
        }

        vrpsx_errno_t setdetachstate_nt(
            int detach_state)
        {
            vrpsx_thread_attr_t *iface = from_cpp(this);
            return iface->vtbl->setdetachstate(iface, detach_state);
        }

        void setdetachstate(
            int detach_state)
        {
            throw_on_error(
                "vrpsx::thread_attr::setdetachstate",
                setdetachstate_nt(detach_state));
        }

        vrpsx_errno_t getscope_nt(
            int& contention_scope) const
        {
            vrpsx_thread_attr_t const *iface = from_cpp(this);
            return iface->vtbl->getscope(iface, &contention_scope);
        }

        int getscope() const
        {
            int result;
            throw_on_error(
                "vrpsx::thread_attr::getscope",
                getscope_nt(result));
            return result;
        }

        vrpsx_errno_t setscope_nt(
            int contention_scope)
        {
            vrpsx_thread_attr_t *iface = from_cpp(this);
            return iface->vtbl->setscope(iface, contention_scope);
        }

        void setscope(
            int contention_scope)
        {
            throw_on_error(
                "vrpsx::thread_attr::setscope",
                setscope_nt(contention_scope));
        }

        vrpsx_errno_t getstacksize_nt(
            std::size_t& stack_size) const
        {
            vrpsx_thread_attr_t const *iface = from_cpp(this);
            return iface->vtbl->getstacksize(iface, &stack_size);
        }

        std::size_t getstacksize() const
        {
            std::size_t result;
            throw_on_error(
                "vrpsx::thread_attr::getstacksize",
                getstacksize_nt(result));
            return result;
        }

        vrpsx_errno_t setstacksize_nt(
            std::size_t stack_size)
        {
            vrpsx_thread_attr_t *iface = from_cpp(this);
            return iface->vtbl->setstacksize(iface, stack_size);
        }

        void setstacksize(
            std::size_t stack_size)
        {
            throw_on_error(
                "vrpsx::thread_attr::setstacksize",
                setstacksize_nt(stack_size));
        }
    };


    VRPSX_CPP_WRAPPER(thread_key_t);

    class thread_key_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_thread_key_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::thread_key::disown",
                sroe, ern);
        }

        void *get() const
        {
            vrpsx_thread_key_t const *iface = from_cpp(this);
            return iface->vtbl->get(iface);
        }

        vrpsx_errno_t set_nt(
            void const *value)
        {
            vrpsx_thread_key_t *iface = from_cpp(this);
            return iface->vtbl->set(iface, value);
        }

        void set(
            void const *value)
        {
            throw_on_error(
                "vrpsx::thread_key::set",
                set_nt(value));
        }
    };


    VRPSX_CPP_WRAPPER(mutex_t);

    class mutex_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_mutex_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::mutex::disown",
                sroe, ern);
        }

        vrpsx_errno_t unlock_nt()
        {
            vrpsx_mutex_t *iface = from_cpp(this);
            return iface->vtbl->unlock(iface);
        }

        void unlock()
        {
            throw_on_error(
                "vrpsx::mutex::unlock",
                unlock_nt());
        }

        vrpsx_errno_t lock_nt()
        {
            vrpsx_mutex_t *iface = from_cpp(this);
            return iface->vtbl->lock(iface);
        }

        void lock()
        {
            throw_on_error(
                "vrpsx::mutex::lock",
                lock_nt());
        }

        vrpsx_errno_t trylock_nt()
        {
            vrpsx_mutex_t *iface = from_cpp(this);
            return iface->vtbl->trylock(iface);
        }

        bool trylock()
        {
            vrpsx_errno_t ern = trylock_nt();
            if (ern) {
                return true;
            } else {
                if (ern != from_cpp(get_system())->vtbl->ebusy) {
                    unconditional_throw("vrpsx::mutex::trylock", ern);
                }
                return false;
            }
        }
    };


    VRPSX_CPP_WRAPPER(mutex_attr_t);

    class mutex_attr_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_mutex_attr_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::mutex_attr::disown",
                sroe, ern);
        }

        vrpsx_errno_t create_nt(
            single_owner_ptr_t< mutex_t >::type& mutex) const
        {
            vrpsx_mutex_attr_t const *iface = from_cpp(this);
            vrpsx_mutex_t *raw_mutex;
            vrpsx_errno_t ern = iface->vtbl->create(
                iface, &raw_mutex);
            to_cpp_single_owner(raw_mutex).swap(mutex);
            return ern;
        }

        single_owner_ptr_t< mutex_t >::type create()
        {
            single_owner_ptr_t< mutex_t >::type result;
            throw_on_error(
                "vrpsx::mutex_attr::create",
                create_nt(result));
            return move(result);
        }

        vrpsx_errno_t gettype_nt(
            int& type)
        {
            vrpsx_mutex_attr_t *iface = from_cpp(this);
            return iface->vtbl->gettype(iface, &type);
        }

        int gettype()
        {
            int type;
            throw_on_error(
                "vrpsx::mutex_attr::gettype",
                gettype_nt(type));
            return type;
        }

        vrpsx_errno_t settype_nt(
            int type)
        {
            vrpsx_mutex_attr_t *iface = from_cpp(this);
            return iface->vtbl->settype(iface, type);
        }

        void settype(
            int type)
        {
            throw_on_error(
                "vrpsx::mutex_attr::settype",
                settype_nt(type));
        }
    };


    // Sentry that locks the given mutex, then unlocks it upon destruction.

    class mutex_lock_t
    {
    public:
        mutex_lock_t(mutex_t& m)
            : mMutex(m)
        {
            m.lock();
        }

        ~mutex_lock_t()
        {
            vrpsx_errno_t ern = mMutex.unlock_nt();
            if (ern && ! std::uncaught_exception()) {
                mMutex.unconditional_throw("vrpsx::mutex::unlock", ern);
            }
        }

        mutex_t& get_mutex() { return mMutex; }

    private:
        mutex_t& mMutex;
    };


    VRPSX_CPP_WRAPPER(rwlock_t);

    class rwlock_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::rwlock::disown",
                sroe, ern);
        }

        vrpsx_errno_t unlock_nt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            return iface->vtbl->unlock(iface);
        }

        void unlock()
        {
            throw_on_error(
                "vrpsx::rwlock::unlock",
                unlock_nt());
        }

        vrpsx_errno_t rdlock_nt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            return iface->vtbl->rdlock(iface);
        }

        void rdlock()
        {
            throw_on_error(
                "vrpsx::rwlock::rdlock",
                rdlock_nt());
        }

        vrpsx_errno_t wrlock_nt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            return iface->vtbl->wrlock(iface);
        }

        void wrlock()
        {
            throw_on_error(
                "vrpsx::rwlock::wrlock",
                wrlock_nt());
        }

        vrpsx_errno_t tryrdlock_nt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            return iface->vtbl->tryrdlock(iface);
        }

        bool tryrdlock()
        {
            vrpsx_errno_t ern = tryrdlock_nt();
            if (ern) {
                return true;
            } else {
                if (ern != from_cpp(get_system())->vtbl->ebusy) {
                    unconditional_throw("vrpsx::rwlock::tryrdlock", ern);
                }
                return false;
            }
        }

        vrpsx_errno_t trywrlock_nt()
        {
            vrpsx_rwlock_t *iface = from_cpp(this);
            return iface->vtbl->trywrlock(iface);
        }

        bool trywrlock()
        {
            vrpsx_errno_t ern = trywrlock_nt();
            if (ern) {
                return true;
            } else {
                if (ern != from_cpp(get_system())->vtbl->ebusy) {
                    unconditional_throw("vrpsx::rwlock::trywrlock", ern);
                }
                return false;
            }
        }
    };


    VRPSX_CPP_WRAPPER(rwlock_attr_t);

    class rwlock_attr_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_rwlock_attr_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::rwlock_attr::disown",
                sroe, ern);
        }

        vrpsx_errno_t create_nt(
            single_owner_ptr_t< rwlock_t >::type& rwlock) const
        {
            vrpsx_rwlock_attr_t const *iface = from_cpp(this);
            vrpsx_rwlock_t *raw_rwlock;
            vrpsx_errno_t ern = iface->vtbl->create(
                iface, &raw_rwlock);
            to_cpp_single_owner(raw_rwlock).swap(rwlock);
            return ern;
        }

        single_owner_ptr_t< rwlock_t >::type create()
        {
            single_owner_ptr_t< rwlock_t >::type result;
            throw_on_error(
                "vrpsx::rwlock_attr::create",
                create_nt(result));
            return move(result);
        }
    };


    VRPSX_CPP_WRAPPER(cond_t);

    class cond_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_cond_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::cond::disown",
                sroe, ern);
        }

        vrpsx_errno_t broadcast_nt()
        {
            vrpsx_cond_t *iface = from_cpp(this);
            return iface->vtbl->broadcast(iface);
        }

        void broadcast()
        {
            throw_on_error(
                "vrpsx::cond::broadcast",
                broadcast_nt());
        }

        vrpsx_errno_t signal_nt()
        {
            vrpsx_cond_t *iface = from_cpp(this);
            return iface->vtbl->signal(iface);
        }

        void signal()
        {
            throw_on_error(
                "vrpsx::cond::signal",
                signal_nt());
        }

        vrpsx_errno_t wait_nt(
            mutex_lock_t& lock)
        {
            vrpsx_cond_t *iface = from_cpp(this);
            return iface->vtbl->wait(
                iface, from_cpp(&lock.get_mutex()));
        }

        void wait(
            mutex_lock_t& lock)
        {
            throw_on_error(
                "vrpsx::cond::wait",
                wait_nt(lock));
        }

        vrpsx_errno_t timedwait_nt(
            mutex_lock_t& lock,
            vrpsx_timespec_t const& absolute_time)
        {
            vrpsx_cond_t *iface = from_cpp(this);
            return iface->vtbl->timedwait(
                iface, from_cpp(&lock.get_mutex()), &absolute_time);
        }

        bool timedwait(
            mutex_lock_t& lock,
            vrpsx_timespec_t const& absolute_time)
        {
            vrpsx_errno_t ern = timedwait_nt(lock, absolute_time);
            if (ern) {
                return true;
            } else {
                if (ern != from_cpp(get_system())->vtbl->etimedout) {
                    unconditional_throw("vrpsx::cond::timedwait", ern);
                }
                return false;
            }
        }
    };


    VRPSX_CPP_WRAPPER(cond_attr_t);

    class cond_attr_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_cond_attr_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::cond_attr::disown",
                sroe, ern);
        }

        vrpsx_errno_t create_nt(
            single_owner_ptr_t< cond_t >::type& cond) const
        {
            vrpsx_cond_attr_t const *iface = from_cpp(this);
            vrpsx_cond_t *raw_cond;
            vrpsx_errno_t ern = iface->vtbl->create(
                iface, &raw_cond);
            to_cpp_single_owner(raw_cond).swap(cond);
            return ern;
        }

        single_owner_ptr_t< cond_t >::type create()
        {
            single_owner_ptr_t< cond_t >::type result;
            throw_on_error(
                "vrpsx::cond_attr::create",
                create_nt(result));
            return move(result);
        }
    };


    VRPSX_CPP_WRAPPER(dirent_t);

    class dirent_t
    {
    public:
        char const *name() const
        {
            vrpsx_dirent_t const *iface = from_cpp(this);
            return iface->vtbl->name(iface);
        }

        vrpsx_ino_t ino() const
        {
            vrpsx_dirent_t const *iface = from_cpp(this);
            return iface->vtbl->ino(iface);
        }
    };


    VRPSX_CPP_WRAPPER(directory_t);

    class directory_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_directory_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->disown(&iface, &sroe);
            disown_check(
                "vrpsx::directory::disown",
                sroe, ern);
        }

        vrpsx_errno_t dirfd_nt(
            vrpsx_fd_t& out_fd)
        {
            vrpsx_directory_t *iface = from_cpp(this);
            return iface->vtbl->dirfd(iface, &out_fd);
        }

        vrpsx_fd_t dirfd(
            vrpsx_directory_t *dir)
        {
            vrpsx_fd_t result;
            throw_on_error(
                "vrpsx::directory::dirfd",
                dirfd_nt(result));
            return result;
        }

        vrpsx_errno_t readdir_nt(
            dirent_t *& next)
        {
            vrpsx_directory_t *iface = from_cpp(this);
            vrpsx_dirent_t *raw_next;
            vrpsx_errno_t ern = iface->vtbl->readdir(iface, &raw_next);
            next = to_cpp(raw_next);
            return ern;
        }

        dirent_t *readdir()
        {
            dirent_t *result;
            throw_on_error(
                "vrpsx::directory::readdir",
                readdir_nt(result));
            return result;
        }

        vrpsx_errno_t rewinddir_nt()
        {
            vrpsx_directory_t *iface = from_cpp(this);
            return iface->vtbl->rewinddir(iface);
        }

        void rewinddir()
        {
            throw_on_error(
                "vrpsx::directory::rewinddir",
                rewinddir_nt());
        }
    };


    VRPSX_CPP_WRAPPER(mount_option_t);

    class mount_option_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_mount_option_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->next(&iface, 0, &sroe);
            disown_check(
                "vrpsx::mount_option::next",
                sroe, ern);
        }

        friend vrpsx_errno_t next_nt(
            single_owner_ptr_t< mount_option_t >::type& option,
            intrusive_ptr< system_t > *sroe)
        {
            vrpsx_mount_option_t *iface = from_cpp(option.release());
            vrpsx_system_t *raw_sroe = 0;
            vrpsx_errno_t ern = iface->vtbl->next(
                &iface, &iface, (sroe ? &raw_sroe : 0));
            option.reset(to_cpp(iface));
            if (sroe) {
                intrusive_ptr< system_t >(to_cpp(raw_sroe), false).swap(*sroe);
            }
            return ern;
        }

        friend void next(
            single_owner_ptr_t< mount_option_t >::type& option)
        {
            intrusive_ptr< system_t > sroe;
            vrpsx_errno_t ern = next_nt(option, &sroe);
            if (ern) {
                vrpsx::unconditional_throw(
                    "vrpsx::mount_option::next",
                    *sroe, ern);
            }
        }

        vrpsx_errno_t text_nt(
            vrpsx_substring_t& text)
        {
            vrpsx_mount_option_t const *iface = from_cpp(this);
            return iface->vtbl->text(iface, &text);
        }

        vrpsx_substring_t text()
        {
            vrpsx_substring_t result;
            throw_on_error(
                "vrpsx::mount_option::text",
                text_nt(result));
            return result;
        }
    };


    VRPSX_CPP_WRAPPER(mount_t);

    class mount_t : public system_associate_t
    {
    public:
        void disown_mt()
        {
            vrpsx_mount_t *iface = from_cpp(this);
            vrpsx_system_t *sroe;
            vrpsx_errno_t ern = iface->vtbl->next(&iface, 0, &sroe);
            disown_check(
                "vrpsx::mount::next",
                sroe, ern);
        }

        friend vrpsx_errno_t next_nt(
            single_owner_ptr_t< mount_t >::type& mnt,
            intrusive_ptr< system_t > *sroe)
        {
            vrpsx_mount_t *iface = from_cpp(mnt.release());
            vrpsx_system_t *raw_sroe = 0;
            vrpsx_errno_t ern = iface->vtbl->next(
                &iface, &iface, (sroe ? &raw_sroe : 0));
            mnt.reset(to_cpp(iface));
            if (sroe) {
                intrusive_ptr< system_t >(to_cpp(raw_sroe), false).swap(*sroe);
            }
            return ern;
        }

        friend void next(
            single_owner_ptr_t< mount_t >::type& mnt)
        {
            intrusive_ptr< system_t > sroe;
            vrpsx_errno_t ern = next_nt(mnt, &sroe);
            if (ern) {
                vrpsx::unconditional_throw(
                    "vrpsx::mount::next",
                    *sroe, ern);
            }
        }

        vrpsx_errno_t device_nt(
            vrpsx_substring_t& device)
        {
            vrpsx_mount_t const *iface = from_cpp(this);
            return iface->vtbl->device(iface, &device);
        }

        vrpsx_substring_t device()
        {
            vrpsx_substring_t result;
            throw_on_error(
                "vrpsx::mount::device",
                device_nt(result));
            return result;
        }

        vrpsx_errno_t root_nt(
            vrpsx_substring_t& root)
        {
            vrpsx_mount_t const *iface = from_cpp(this);
            return iface->vtbl->root(iface, &root);
        }

        vrpsx_substring_t root()
        {
            vrpsx_substring_t result;
            throw_on_error(
                "vrpsx::mount::root",
                root_nt(result));
            return result;
        }

        vrpsx_errno_t type_nt(
            vrpsx_substring_t& type)
        {
            vrpsx_mount_t const *iface = from_cpp(this);
            return iface->vtbl->type(iface, &type);
        }

        vrpsx_substring_t type()
        {
            vrpsx_substring_t result;
            throw_on_error(
                "vrpsx::mount::type",
                type_nt(result));
            return result;
        }

        // WARNING: Be sure to clear "options" before calling this
        // method again or disowning this object, in order to avoid
        // double-disowning the option sequence.

        vrpsx_errno_t options_nt(
            single_owner_ptr_t< mount_option_t >::type& options)
        {
            vrpsx_mount_t *iface = from_cpp(this);
            vrpsx_mount_option_t *raw_options;
            vrpsx_errno_t ern = iface->vtbl->options(
                iface, &raw_options);
            to_cpp_single_owner(raw_options).swap(options);
            return ern;
        }

        // WARNING: See the warning for "options_nt".

        single_owner_ptr_t< mount_option_t >::type options()
        {
            single_owner_ptr_t< mount_option_t >::type result;
            throw_on_error(
                "vrpsx::mount::options",
                options_nt(result));
            return move(result);
        }
    };


    class system_allocator_padding_t
    {
    private:
        char mPadding[offsetof(vrpsx_system_t, allocator)];
    };

    class system_t :
        public counted_identified_t,
        private system_allocator_padding_t,
        public allocator_t  // because allocator_t has nonzero size,
                            // even when used as a base class, the
                            // above padding ensures the correct offset
    {
    public:
        void unconditional_throw(
            const char *prefix,
            vrpsx_errno_t ern)
        {
            vrpsx::unconditional_throw(prefix, *this, ern);
        }

        void throw_on_error(
            const char *prefix,
            vrpsx_errno_t ern)
        {
            if (ern) {
                unconditional_throw(prefix, ern);
            }
        }

        template< typename M >
        unique_ptr< M, free_memory_t > unique_allocation(
            M *memory,
            allocator_t *allocator)
        {
            if (! allocator) {
                allocator = this;
            }

            return unique_ptr< M, free_memory_t >(
                memory, free_memory_t(allocator));
        }

        intrusive_ptr< system_error_t > error_acquire(
            vrpsx_errno_t ern)
        {
            return vrpsx::error_acquire(*this, ern);
        }

        void exit(
            int code)
        {
            vrpsx_system_t *iface = from_cpp(this);
            iface->vtbl->exit(iface, code);
        }

        vrpsx_errno_t uname_nt(
            single_owner_ptr_t< utsname_t >::type& out_name)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_utsname_t *raw_out_name;
            vrpsx_errno_t ern = iface->vtbl->uname(
                iface, &raw_out_name);
            to_cpp_single_owner(raw_out_name).swap(out_name);
            return ern;
        }

        single_owner_ptr_t< utsname_t >::type uname()
        {
            single_owner_ptr_t< utsname_t >::type result;
            throw_on_error(
                "vrpsx::system::uname",
                uname_nt(result));
            return move(result);
        }

        vrpsx_errno_t sysconf_nt(
            int name,
            long& out_value)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->sysconf(iface, name, &out_value);
        }

        long sysconf(
            int name)
        {
            long result;
            throw_on_error(
                "vrpsx::system::sysconf",
                sysconf_nt(name, result));
            return result;
        }

        vrpsx_errno_t pathconf_nt(
            char const *path,
            int name,
            long& out_value)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->pathconf(
                iface, path, name, &out_value);
        }

        long pathconf(
            char const *path,
            int name)
        {
            long result;
            throw_on_error(
                "vrpsx::system::pathconf",
                pathconf_nt(path, name, result));
            return result;
        }

        vrpsx_errno_t fpathconf_nt(
            vrpsx_fd_t fd,
            int name,
            long& out_value)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->fpathconf(
                iface, fd, name, &out_value);
        }

        long fpathconf(
            vrpsx_fd_t fd,
            int name)
        {
            long result;
            throw_on_error(
                "vrpsx::system::fpathconf",
                fpathconf_nt(fd, name, result));
            return result;
        }

        // NOTE: Be sure to destroy or release out_cwd before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t getcwd_nt(
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& out_cwd)
        {
            vrpsx_system_t *iface = from_cpp(this);
            char *raw_out_cwd;
            vrpsx_errno_t ern = iface->vtbl->getcwd(
                iface, from_cpp(allocator), &raw_out_cwd);
            unique_allocation(raw_out_cwd, allocator).swap(out_cwd);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > getcwd(
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::system::getcwd",
                getcwd_nt(allocator, result));
            return move(result);
        }

        vrpsx_int64_t get_timezone()
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->get_timezone(iface);
        }

        char const *const *get_tzname()
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->get_tzname(iface);
        }

        vrpsx_errno_t gmtime_nt(
            vrpsx_time_t in_time,
            vrpsx_tm_t& out_time)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->gmtime(iface, in_time, &out_time);
        }

        vrpsx_tm_t gmtime(
            vrpsx_time_t in_time)
        {
            vrpsx_tm_t result;
            throw_on_error(
                "vrpsx::system::gmtime",
                gmtime_nt(in_time, result));
            return result;
        }

        vrpsx_errno_t localtime_nt(
            vrpsx_time_t in_time,
            vrpsx_tm_t& out_time)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->localtime(iface, in_time, &out_time);
        }

        vrpsx_tm_t localtime(
            vrpsx_time_t in_time)
        {
            vrpsx_tm_t result;
            throw_on_error(
                "vrpsx::system::localtime",
                localtime_nt(in_time, result));
            return result;
        }

        vrpsx_errno_t mktime_nt(
            vrpsx_tm_t& io_tm,
            vrpsx_time_t& out_time)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->mktime(iface, &io_tm, &out_time);
        }

        vrpsx_time_t mktime(
            vrpsx_tm_t& io_tm)
        {
            vrpsx_time_t result;
            throw_on_error(
                "vrpsx::system::mktime",
                mktime_nt(io_tm, result));
            return result;
        }

        vrpsx_errno_t time_nt(
            vrpsx_time_t& out_time)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->time(iface, &out_time);
        }

        vrpsx_time_t time()
        {
            vrpsx_time_t result;
            throw_on_error(
                "vrpsx::system::time",
                time_nt(result));
            return result;
        }

        vrpsx_errno_t clock_gettime_nt(
            vrpsx_clockid_t clock,
            vrpsx_timespec_t& out_time)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->clock_gettime(iface, clock, &out_time);
        }

        vrpsx_timespec_t clock_gettime(
            vrpsx_clockid_t clock)
        {
            vrpsx_timespec_t result;
            throw_on_error(
                "vrpsx::system::clock_gettime",
                clock_gettime_nt(clock, result));
            return result;
        }

        vrpsx_errno_t nanosleep_nt(
            vrpsx_timespec_t const& duration,
            vrpsx_timespec_t *remaining)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->nanosleep(iface, &duration, remaining);
        }

        vrpsx_timespec_t nanosleep(
            vrpsx_timespec_t const& duration)
        {
            vrpsx_timespec_t result;
            throw_on_error(
                "vrpsx::system::nanosleep",
                nanosleep_nt(duration, &result));
            return result;
        }

        // NOTE: Be sure to destroy or release out_name before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t getlogin_nt(
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& out_name)
        {
            vrpsx_system_t *iface = from_cpp(this);
            char *raw_out_name;
            vrpsx_errno_t ern = iface->vtbl->getlogin(
                iface, from_cpp(allocator), &raw_out_name);
            unique_allocation(raw_out_name, allocator).swap(out_name);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > getlogin(
            allocator_t *allocator)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::system::getlogin",
                getlogin_nt(allocator, result));
            return move(result);
        }

        vrpsx_errno_t getpwnam_nt(
            char const *name,
            single_owner_ptr_t< passwd_t >::type& result)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_passwd_t *raw_result;
            vrpsx_errno_t ern = iface->vtbl->getpwnam(
                iface, name, &raw_result);
            to_cpp_single_owner(raw_result).swap(result);
            return ern;
        }

        single_owner_ptr_t< passwd_t >::type getpwnam(
            char const *name)
        {
            single_owner_ptr_t< passwd_t >::type result;
            throw_on_error(
                "vrpsx::system::getpwnam",
                getpwnam_nt(name, result));
            return move(result);
        }

        vrpsx_pid_t getpid()
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->getpid(iface);
        }

        vrpsx_pid_t getpgrp()
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->getpgrp(iface);
        }

        vrpsx_errno_t getpgid_nt(
            vrpsx_pid_t pid,
            vrpsx_pid_t& out_pgid)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->getpgid(iface, pid, &out_pgid);
        }

        vrpsx_pid_t getpgid(
            vrpsx_pid_t pid)
        {
            vrpsx_pid_t result;
            throw_on_error(
                "vrpsx::system::getpgid",
                getpgid_nt(pid, result));
            return result;
        }

        vrpsx_errno_t kill_nt(
            vrpsx_pid_t pid,
            vrpsx_signo_t signo)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->kill(iface, pid, signo);
        }

        void kill(
            vrpsx_pid_t pid,
            vrpsx_signo_t signo)
        {
            throw_on_error(
                "vrpsx::system::kill",
                kill_nt(pid, signo));
        }

        vrpsx_errno_t raise_nt(
            vrpsx_signo_t signo)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->raise(iface, signo);
        }

        void raise(
            vrpsx_signo_t signo)
        {
            throw_on_error(
                "vrpsx::system::raise",
                raise_nt(signo));
        }

        char const *strsignal(
            vrpsx_signo_t signo)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->strsignal(iface, signo);
        }

        vrpsx_errno_t sigset_init_nt(
            void *storage,
            sigset_t *& set)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_sigset_t *raw_set;
            vrpsx_errno_t ern = iface->vtbl->sigset_init(
                iface, storage, &raw_set);
            set = to_cpp(raw_set);
            return ern;
        }

        sigset_t *sigset_init(
            void *storage)
        {
            sigset_t *result;
            throw_on_error(
                "vrpsx::system::sigset_init",
                sigset_init_nt(storage, result));
            return result;
        }

        vrpsx_errno_t sigaction_init_nt(
            void *storage,
            sigaction_t *& action)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_sigaction_t *raw_action;
            vrpsx_errno_t ern = iface->vtbl->sigaction_init(
                iface, storage, &raw_action);
            action = to_cpp(raw_action);
            return ern;
        }

        sigaction_t *sigaction_init(
            void *storage)
        {
            sigaction_t *result;
            throw_on_error(
                "vrpsx::system::sigaction_init",
                sigaction_init_nt(storage, result));
            return result;
        }

        vrpsx_errno_t sigaction_nt(
            vrpsx_signo_t signo,
            sigaction_t const& action,
            sigaction_t *old_action)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->sigaction(
                iface, signo, from_cpp(&action), from_cpp(old_action));
        }

        void sigaction(
            vrpsx_signo_t signo,
            sigaction_t const& action,
            sigaction_t *old_action)
        {
            throw_on_error(
                "vrpsx::system::sigaction",
                sigaction_nt(signo, action, old_action));
        }

        vrpsx_errno_t sigwait_nt(
            sigset_t const& set,
            vrpsx_signo_t& signo)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->sigwait(
                iface, from_cpp(&set), &signo);
        }

        vrpsx_signo_t sigwait(
            sigset_t const& set)
        {
            vrpsx_signo_t result;
            throw_on_error(
                "vrpsx::system::sigwait",
                sigwait_nt(set, result));
            return result;
        }

        vrpsx_errno_t thread_sigmask_nt(
            int how,
            sigset_t const& change,
            sigset_t *old_set)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_sigmask(
                iface, how, from_cpp(&change), from_cpp(old_set));
        }

        void thread_sigmask(
            int how,
            sigset_t const& change,
            sigset_t *old_set)
        {
            throw_on_error(
                "vrpsx::system::thread_sigmask",
                thread_sigmask_nt(how, change, old_set));
        }

        vrpsx_errno_t thread_attr_create_nt(
            single_owner_ptr_t< thread_attr_t >::type& attr)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_thread_attr_t *raw_attr;
            vrpsx_errno_t ern = iface->vtbl->thread_attr_create(
                iface, &raw_attr);
            to_cpp_single_owner(raw_attr).swap(attr);
            return ern;
        }

        single_owner_ptr_t< thread_attr_t >::type thread_attr_create()
        {
            single_owner_ptr_t< thread_attr_t >::type result;
            throw_on_error(
                "vrpsx::system::thread_attr_create",
                thread_attr_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t thread_create_nt(
            vrpsx_thread_t& thread,
            vrpsx_thread_function_t *function,
            void *argument)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_create(
                iface, &thread, function, argument);
        }

        vrpsx_thread_t thread_create(
            vrpsx_thread_function_t *function,
            void *argument)
        {
            vrpsx_thread_t result;
            throw_on_error(
                "vrpsx::system::thread_create",
                thread_create_nt(result, function, argument));
            return result;
        }

        vrpsx_errno_t thread_detach_nt(
            vrpsx_thread_t thread)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_detach(iface, thread);
        }

        void thread_detach(
            vrpsx_thread_t thread)
        {
            throw_on_error(
                "vrpsx::system::thread_detach",
                thread_detach_nt(thread));
        }

        vrpsx_errno_t thread_join_nt(
            vrpsx_thread_t thread,
            void **value)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_join(iface, thread, value);
        }

        void *thread_join(
            vrpsx_thread_t thread)
        {
            void *result;
            throw_on_error(
                "vrpsx::system::thread_join",
                thread_join_nt(thread, &result));
            return result;
        }

        vrpsx_errno_t thread_kill_nt(
            vrpsx_thread_t thread,
            vrpsx_signo_t signo)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_kill(iface, thread, signo);
        }

        void thread_kill(
            vrpsx_thread_t thread,
            vrpsx_signo_t signo)
        {
            throw_on_error(
                "vrpsx::system::thread_kill",
                thread_kill_nt(thread, signo));
        }

        bool thread_equal(
            vrpsx_thread_t thread1,
            vrpsx_thread_t thread2)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return 0 != iface->vtbl->thread_equal(
                iface, thread1, thread2);
        }

        vrpsx_thread_t thread_self()
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->thread_self(iface);
        }

        // WARNING: See the comments at the definition of
        // vrpsx_thread_key_t for the definitive comments, but:
        // do not destroy the smart pointer that currently owns
        // a given key until all threads stop using that key.

        vrpsx_errno_t thread_key_create_nt(
            single_owner_ptr_t< thread_key_t >::type& key,
            vrpsx_thread_key_destructor_t *destructor)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_thread_key_t *raw_key;
            vrpsx_errno_t ern = iface->vtbl->thread_key_create(
                iface, &raw_key, destructor);
            to_cpp_single_owner(raw_key).swap(key);
            return ern;
        }

        // WARNING: See the comments at the definition of
        // vrpsx_thread_key_t for the definitive comments, but:
        // do not destroy the smart pointer that currently owns
        // a given key until all threads stop using that key.

        single_owner_ptr_t< thread_key_t >::type thread_key_create(
            vrpsx_thread_key_destructor_t *destructor)
        {
            single_owner_ptr_t< thread_key_t >::type result;
            throw_on_error(
                "vrpsx::system::thread_key_create",
                thread_key_create_nt(result, destructor));
            return move(result);
        }

        vrpsx_errno_t mutex_attr_create_nt(
            single_owner_ptr_t< mutex_attr_t >::type& attr)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_mutex_attr_t *raw_attr;
            vrpsx_errno_t ern = iface->vtbl->mutex_attr_create(
                iface, &raw_attr);
            to_cpp_single_owner(raw_attr).swap(attr);
            return ern;
        }

        single_owner_ptr_t< mutex_attr_t >::type mutex_attr_create()
        {
            single_owner_ptr_t< mutex_attr_t >::type result;
            throw_on_error(
                "vrpsx::system::mutex_attr_create",
                mutex_attr_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t mutex_create_nt(
            single_owner_ptr_t< mutex_t >::type& mutex)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_mutex_t *raw_mutex;
            vrpsx_errno_t ern = iface->vtbl->mutex_create(
                iface, &raw_mutex);
            to_cpp_single_owner(raw_mutex).swap(mutex);
            return ern;
        }

        single_owner_ptr_t< mutex_t >::type mutex_create()
        {
            single_owner_ptr_t< mutex_t >::type result;
            throw_on_error(
                "vrpsx::system::mutex_create",
                mutex_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t rwlock_attr_create_nt(
            single_owner_ptr_t< rwlock_attr_t >::type& attr)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_rwlock_attr_t *raw_attr;
            vrpsx_errno_t ern = iface->vtbl->rwlock_attr_create(
                iface, &raw_attr);
            to_cpp_single_owner(raw_attr).swap(attr);
            return ern;
        }

        single_owner_ptr_t< rwlock_attr_t >::type rwlock_attr_create()
        {
            single_owner_ptr_t< rwlock_attr_t >::type result;
            throw_on_error(
                "vrpsx::system::rwlock_attr_create",
                rwlock_attr_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t rwlock_create_nt(
            single_owner_ptr_t< rwlock_t >::type& rwlock)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_rwlock_t *raw_rwlock;
            vrpsx_errno_t ern = iface->vtbl->rwlock_create(
                iface, &raw_rwlock);
            to_cpp_single_owner(raw_rwlock).swap(rwlock);
            return ern;
        }

        single_owner_ptr_t< rwlock_t >::type rwlock_create()
        {
            single_owner_ptr_t< rwlock_t >::type result;
            throw_on_error(
                "vrpsx::system::rwlock_create",
                rwlock_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t cond_attr_create_nt(
            single_owner_ptr_t< cond_attr_t >::type& attr)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_cond_attr_t *raw_attr;
            vrpsx_errno_t ern = iface->vtbl->cond_attr_create(
                iface, &raw_attr);
            to_cpp_single_owner(raw_attr).swap(attr);
            return ern;
        }

        single_owner_ptr_t< cond_attr_t >::type cond_attr_create()
        {
            single_owner_ptr_t< cond_attr_t >::type result;
            throw_on_error(
                "vrpsx::system::cond_attr_create",
                cond_attr_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t cond_create_nt(
            single_owner_ptr_t< cond_t >::type& cond)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_cond_t *raw_cond;
            vrpsx_errno_t ern = iface->vtbl->cond_create(
                iface, &raw_cond);
            to_cpp_single_owner(raw_cond).swap(cond);
            return ern;
        }

        single_owner_ptr_t< cond_t >::type cond_create()
        {
            single_owner_ptr_t< cond_t >::type result;
            throw_on_error(
                "vrpsx::system::cond_create",
                cond_create_nt(result));
            return move(result);
        }

        vrpsx_errno_t open_nt(
            char const *path,
            int flags,
            int permissions,
            vrpsx_fd_t& out_fd)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->open(
                iface, path, flags, permissions, &out_fd);
        }

        vrpsx_fd_t open(
            char const *path,
            int flags,
            int permissions)
        {
            vrpsx_fd_t result;
            throw_on_error(
                "vrpsx::system::open",
                open_nt(path, flags, permissions, result));
            return result;
        }

        vrpsx_errno_t close_nt(
            vrpsx_fd_t fd)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->close(iface, fd);
        }

        void close(
            vrpsx_fd_t fd)
        {
            throw_on_error(
                "vrpsx::system::close",
                close_nt(fd));
        }

        vrpsx_errno_t read_nt(
            vrpsx_fd_t fd,
            void *buffer,
            std::size_t in_size,
            std::size_t& out_size)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->read(
                iface, fd, buffer, in_size, &out_size);
        }

        std::size_t read(
            vrpsx_fd_t fd,
            void *buffer,
            std::size_t in_size)
        {
            std::size_t result;
            throw_on_error(
                "vrpsx::system::read",
                read_nt(fd, buffer, in_size, result));
            return result;
        }

        vrpsx_errno_t write_nt(
            vrpsx_fd_t fd,
            void const *buffer,
            std::size_t in_size,
            std::size_t *out_size)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->write(
                iface, fd, buffer, in_size, out_size);
        }

        std::size_t write(
            vrpsx_fd_t fd,
            void const *buffer,
            std::size_t in_size)
        {
            std::size_t result;
            throw_on_error(
                "vrpsx::system::write",
                write_nt(fd, buffer, in_size, &result));
            return result;
        }

        vrpsx_errno_t pread_nt(
            vrpsx_fd_t fd,
            void *buffer,
            std::size_t in_size,
            std::size_t offset,
            std::size_t& out_size)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->pread(
                iface, fd, buffer, in_size, offset, &out_size);
        }

        std::size_t pread(
            vrpsx_fd_t fd,
            void *buffer,
            std::size_t in_size,
            std::size_t offset)
        {
            std::size_t result;
            throw_on_error(
                "vrpsx::system::pread",
                pread_nt(fd, buffer, in_size, offset, result));
            return result;
        }

        vrpsx_errno_t pwrite_nt(
            vrpsx_fd_t fd,
            void const *buffer,
            std::size_t in_size,
            std::size_t offset,
            std::size_t *out_size)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->pwrite(
                iface, fd, buffer, in_size, offset, out_size);
        }

        std::size_t pwrite(
            vrpsx_fd_t fd,
            void const *buffer,
            std::size_t in_size,
            std::size_t offset)
        {
            std::size_t result;
            throw_on_error(
                "vrpsx::system::pwrite",
                pwrite_nt(fd, buffer, in_size, offset, &result));
            return result;
        }

        vrpsx_errno_t unlink_nt(
            char const *path)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->unlink(iface, path);
        }

        void unlink(
            char const *path)
        {
            throw_on_error(
                "vrpsx::system::unlink",
                unlink_nt(path));
        }

        vrpsx_errno_t link_nt(
            char const *existing,
            char const *new_name)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->link(iface, existing, new_name);
        }

        void link(
            char const *existing,
            char const *new_name)
        {
            throw_on_error(
                "vrpsx::system::link",
                link_nt(existing, new_name));
        }

        vrpsx_errno_t symlink_nt(
            char const *target,
            char const *symlink_name)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->symlink(iface, target, symlink_name);
        }

        void symlink(
            char const *target,
            char const *symlink_name)
        {
            throw_on_error(
                "vrpsx::system::symlink",
                symlink_nt(target, symlink_name));
        }

        // NOTE: Be sure to destroy or release result before *allocator
        // (or before releasing this interface, if allocator == NULL).

        vrpsx_errno_t readlink_nt(
            char const *path,
            allocator_t *allocator,
            unique_ptr< char, free_memory_t >& result,
            std::size_t& result_length)
        {
            vrpsx_system_t *iface = from_cpp(this);
            char *raw_result;
            vrpsx_errno_t ern = iface->vtbl->readlink(
                iface, path, from_cpp(allocator),
                &raw_result, &result_length);
            unique_allocation(raw_result, allocator).swap(result);
            return ern;
        }

        // NOTE: Be sure to destroy or release the return before *allocator
        // (or before releasing this interface, if allocator == NULL).

        unique_ptr< char, free_memory_t > readlink(
            char const *path,
            allocator_t *alloc,
            std::size_t& result_length)
        {
            unique_ptr< char, free_memory_t > result;
            throw_on_error(
                "vrpsx::system::readlink",
                readlink_nt(path, alloc, result, result_length));
            return move(result);
        }

        vrpsx_errno_t chmod_nt(
            char const *path,
            vrpsx_mode_t mode)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->chmod(iface, path, mode);
        }

        void chmod(
            char const *path,
            vrpsx_mode_t mode)
        {
            throw_on_error(
                "vrpsx::system::chmod",
                chmod_nt(path, mode));
        }

        vrpsx_errno_t fchmod_nt(
            vrpsx_fd_t fd,
            vrpsx_mode_t mode)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->fchmod(iface, fd, mode);
        }

        void fchmod(
            vrpsx_fd_t fd,
            vrpsx_mode_t mode)
        {
            throw_on_error(
                "vrpsx::system::fchmod",
                fchmod_nt(fd, mode));
        }

        vrpsx_errno_t stat_nt(
            char const *path,
            vrpsx_stat_t& buffer)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->stat(iface, path, &buffer);
        }

        vrpsx_stat_t stat(
            char const *path)
        {
            vrpsx_stat_t result;
            throw_on_error(
                "vrpsx::system::stat",
                stat_nt(path, result));
            return result;
        }

        vrpsx_errno_t lstat_nt(
            char const *path,
            vrpsx_stat_t& buffer)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->lstat(iface, path, &buffer);
        }

        vrpsx_stat_t lstat(
            char const *path)
        {
            vrpsx_stat_t result;
            throw_on_error(
                "vrpsx::system::lstat",
                lstat_nt(path, result));
            return result;
        }

        vrpsx_errno_t fstat_nt(
            vrpsx_fd_t fd,
            vrpsx_stat_t& buffer)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->fstat(iface, fd, &buffer);
        }

        vrpsx_stat_t fstat(
            vrpsx_fd_t fd)
        {
            vrpsx_stat_t result;
            throw_on_error(
                "vrpsx::system::fstat",
                fstat_nt(fd, result));
            return result;
        }

        vrpsx_errno_t mkdir_nt(
            char const *path,
            vrpsx_mode_t mode)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->mkdir(iface, path, mode);
        }

        void mkdir(
            char const *path,
            vrpsx_mode_t mode)
        {
            throw_on_error(
                "vrpsx::system::mkdir",
                mkdir_nt(path, mode));
        }

        vrpsx_errno_t rmdir_nt(
            char const *path)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->rmdir(iface, path);
        }

        void rmdir(
            char const *path)
        {
            throw_on_error(
                "vrpsx::system::rmdir",
                rmdir_nt(path));
        }

        vrpsx_errno_t chdir_nt(
            char const *path)
        {
            vrpsx_system_t *iface = from_cpp(this);
            return iface->vtbl->chdir(iface, path);
        }

        void chdir(
            char const *path)
        {
            throw_on_error(
                "vrpsx::system::chdir",
                chdir_nt(path));
        }

        vrpsx_errno_t opendir_nt(
            char const *path,
            single_owner_ptr_t< directory_t >::type& dir)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_directory_t *raw_dir;
            vrpsx_errno_t ern = iface->vtbl->opendir(
                iface, path, &raw_dir);
            to_cpp_single_owner(raw_dir).swap(dir);
            return ern;
        }

        single_owner_ptr_t< directory_t >::type opendir(
            char const *path)
        {
            single_owner_ptr_t< directory_t >::type result;
            throw_on_error(
                "vrpsx::system::opendir",
                opendir_nt(path, result));
            return move(result);
        }

        vrpsx_errno_t fdopendir_nt(
            vrpsx_fd_t fd,
            single_owner_ptr_t< directory_t >::type& dir)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_directory_t *raw_dir;
            vrpsx_errno_t ern = iface->vtbl->fdopendir(
                iface, fd, &raw_dir);
            to_cpp_single_owner(raw_dir).swap(dir);
            return ern;
        }

        single_owner_ptr_t< directory_t >::type fdopendir(
            vrpsx_fd_t fd)
        {
            single_owner_ptr_t< directory_t >::type result;
            throw_on_error(
                "vrpsx::system::fdopendir",
                fdopendir_nt(fd, result));
            return move(result);
        }

        vrpsx_errno_t get_mounts_nt(
            single_owner_ptr_t< mount_t >::type& mounts)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_mount_t *raw_mounts;
            vrpsx_errno_t ern = iface->vtbl->get_mounts(
                iface, &raw_mounts);
            to_cpp_single_owner(raw_mounts).swap(mounts);
            return ern;
        }

        single_owner_ptr_t< mount_t >::type get_mounts()
        {
            single_owner_ptr_t< mount_t >::type result;
            throw_on_error(
                "vrpsx::system::get_mounts",
                get_mounts_nt(result));
            return move(result);
        }

        vrpsx_errno_t mount_nt(
            char const *device,
            char const *root,
            char const *type,
            single_owner_ptr_t< mount_option_t >::type& options)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_mount_option_t *raw_options = from_cpp(options.release());
            return iface->vtbl->mount(
                iface, device, root, type, &raw_options);
        }

        void mount(
            char const *device,
            char const *root,
            char const *type,
            single_owner_ptr_t< mount_option_t >::type& options)
        {
            throw_on_error(
                "vrpsx::system::mount",
                mount_nt(device, root, type, options));
        }

        vrpsx_errno_t umount_nt(
            char const *device_or_root,
            single_owner_ptr_t< mount_option_t >::type& options)
        {
            vrpsx_system_t *iface = from_cpp(this);
            vrpsx_mount_option_t *raw_options = from_cpp(options.release());
            return iface->vtbl->umount(
                iface, device_or_root, &raw_options);
        }

        void umount(
            char const *device_or_root,
            single_owner_ptr_t< mount_option_t >::type& options)
        {
            throw_on_error(
                "vrpsx::system::umount",
                umount_nt(device_or_root, options));
        }

        VRPSX_CONSTANT_METHOD(int, exit_failure)
        VRPSX_CONSTANT_METHOD(int, exit_success)

        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, e2big)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eacces)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eaddrinuse)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eaddrnotavail)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eafnosupport)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eagain)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ealready)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ebadf)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ebadmsg)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ebusy)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ecanceled)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, echild)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, econnaborted)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, econnrefused)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, econnreset)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, edeadlk)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, edestaddrreq)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, edom)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, edquot)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eexist)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, efault)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, efbig)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ehostunreach)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eidrm)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eilseq)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, einprogress)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eintr)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, einval)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eio)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eisconn)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eisdir)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eloop)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, emfile)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, emlink)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, emsgsize)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, emultihop)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enametoolong)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enetdown)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enetunreach)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enfile)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enobufs)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enodata)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enodev)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enoent)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enoexec)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enolck)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enolink)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enomedium)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enomem)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enomsg)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enoprotoopt)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enospc)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enosr)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enostr)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enosys)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotconn)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotdir)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotempty)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotsock)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotsup)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enotty)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, enxio)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eopnotsupp)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eoverflow)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eperm)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, epipe)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eproto)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eprotonosupport)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, eprototype)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, erange)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, erofs)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, espipe)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, esrch)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, estale)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, etime)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, etimedout)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, etxtbsy)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, ewouldblock)
        VRPSX_CONSTANT_METHOD(vrpsx_errno_t, exdev)

        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigabrt)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigalrm)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigbus)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigchld)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigcont)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigfpe)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sighup)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigill)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigint)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigkill)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigpipe)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigpoll)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigprof)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigquit)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigsegv)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigstop)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigsys)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigterm)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigtrap)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigtstp)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigttin)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigttou)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigurg)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigusr1)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigusr2)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigvtalrm)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigxcpu)
        VRPSX_CONSTANT_METHOD(vrpsx_signo_t, sigxfsz)

        VRPSX_CONSTANT_METHOD(vrpsx_sa_handler_t *, sig_dfl)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_handler_t *, sig_err)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_handler_t *, sig_ign)

        VRPSX_CONSTANT_METHOD(std::size_t, sizeof_sigset)
        VRPSX_CONSTANT_METHOD(std::size_t, alignof_sigset)
        VRPSX_CONSTANT_METHOD(std::size_t, sizeof_sigaction)
        VRPSX_CONSTANT_METHOD(std::size_t, alignof_sigaction)

        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_nocldstop)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_nocldwait)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_nodefer)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_onstack)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_resethand)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_restart)
        VRPSX_CONSTANT_METHOD(vrpsx_sa_flags_t, sa_siginfo)
        VRPSX_CONSTANT_METHOD(int, sig_block)
        VRPSX_CONSTANT_METHOD(int, sig_setmask)
        VRPSX_CONSTANT_METHOD(int, sig_unblock)

        VRPSX_CONSTANT_METHOD(int, thread_create_detached)
        VRPSX_CONSTANT_METHOD(int, thread_create_joinable)
        VRPSX_CONSTANT_METHOD(int, thread_mutex_default)
        VRPSX_CONSTANT_METHOD(int, thread_mutex_errorcheck)
        VRPSX_CONSTANT_METHOD(int, thread_mutex_normal)
        VRPSX_CONSTANT_METHOD(int, thread_mutex_recursive)
        VRPSX_CONSTANT_METHOD(int, thread_process_private)
        VRPSX_CONSTANT_METHOD(int, thread_process_shared)
        VRPSX_CONSTANT_METHOD(int, thread_scope_process)
        VRPSX_CONSTANT_METHOD(int, thread_scope_system)

        VRPSX_CONSTANT_METHOD(int, f_dupfd)
        VRPSX_CONSTANT_METHOD(int, f_getfd)
        VRPSX_CONSTANT_METHOD(int, f_getfl)
        VRPSX_CONSTANT_METHOD(int, f_getlk)
        VRPSX_CONSTANT_METHOD(int, f_rdlck)
        VRPSX_CONSTANT_METHOD(int, f_setfd)
        VRPSX_CONSTANT_METHOD(int, f_setfl)
        VRPSX_CONSTANT_METHOD(int, f_setlk)
        VRPSX_CONSTANT_METHOD(int, f_setlkw)
        VRPSX_CONSTANT_METHOD(int, f_unlck)
        VRPSX_CONSTANT_METHOD(int, f_wrlck)
        VRPSX_CONSTANT_METHOD(int, fd_cloexec)
        VRPSX_CONSTANT_METHOD(int, o_accmode)
        VRPSX_CONSTANT_METHOD(int, o_append)
        VRPSX_CONSTANT_METHOD(int, o_creat)
        VRPSX_CONSTANT_METHOD(int, o_dsync)
        VRPSX_CONSTANT_METHOD(int, o_excl)
        VRPSX_CONSTANT_METHOD(int, o_noctty)
        VRPSX_CONSTANT_METHOD(int, o_nonblock)
        VRPSX_CONSTANT_METHOD(int, o_rdonly)
        VRPSX_CONSTANT_METHOD(int, o_rdwr)
        VRPSX_CONSTANT_METHOD(int, o_rsync)
        VRPSX_CONSTANT_METHOD(int, o_sync)
        VRPSX_CONSTANT_METHOD(int, o_trunc)
        VRPSX_CONSTANT_METHOD(int, o_wronly)

        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifmt)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifblk)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifchr)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ififo)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifreg)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifdir)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_iflnk)
        VRPSX_CONSTANT_METHOD(vrpsx_mode_t, s_ifsock)

        VRPSX_CONSTANT_METHOD(vrpsx_clockid_t, clock_monotonic)
        VRPSX_CONSTANT_METHOD(vrpsx_clockid_t, clock_realtime)

        VRPSX_CONSTANT_METHOD(std::size_t, name_max)
        VRPSX_CONSTANT_METHOD(std::size_t, path_max)

        VRPSX_CONSTANT_METHOD(int, _sc_2_c_bind)
        VRPSX_CONSTANT_METHOD(int, _sc_2_c_dev)
        VRPSX_CONSTANT_METHOD(int, _sc_2_char_term)
        VRPSX_CONSTANT_METHOD(int, _sc_2_fort_dev)
        VRPSX_CONSTANT_METHOD(int, _sc_2_fort_run)
        VRPSX_CONSTANT_METHOD(int, _sc_2_localedef)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs_accounting)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs_checkpoint)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs_locate)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs_message)
        VRPSX_CONSTANT_METHOD(int, _sc_2_pbs_track)
        VRPSX_CONSTANT_METHOD(int, _sc_2_sw_dev)
        VRPSX_CONSTANT_METHOD(int, _sc_2_upe)
        VRPSX_CONSTANT_METHOD(int, _sc_2_version)
        VRPSX_CONSTANT_METHOD(int, _sc_advisory_info)
        VRPSX_CONSTANT_METHOD(int, _sc_aio_listio_max)
        VRPSX_CONSTANT_METHOD(int, _sc_aio_max)
        VRPSX_CONSTANT_METHOD(int, _sc_aio_prio_delta_max)
        VRPSX_CONSTANT_METHOD(int, _sc_arg_max)
        VRPSX_CONSTANT_METHOD(int, _sc_asynchronous_io)
        VRPSX_CONSTANT_METHOD(int, _sc_atexit_max)
        VRPSX_CONSTANT_METHOD(int, _sc_barriers)
        VRPSX_CONSTANT_METHOD(int, _sc_bc_base_max)
        VRPSX_CONSTANT_METHOD(int, _sc_bc_dim_max)
        VRPSX_CONSTANT_METHOD(int, _sc_bc_scale_max)
        VRPSX_CONSTANT_METHOD(int, _sc_bc_string_max)
        VRPSX_CONSTANT_METHOD(int, _sc_child_max)
        VRPSX_CONSTANT_METHOD(int, _sc_clk_tck)
        VRPSX_CONSTANT_METHOD(int, _sc_clock_selection)
        VRPSX_CONSTANT_METHOD(int, _sc_coll_weights_max)
        VRPSX_CONSTANT_METHOD(int, _sc_cputime)
        VRPSX_CONSTANT_METHOD(int, _sc_delaytimer_max)
        VRPSX_CONSTANT_METHOD(int, _sc_expr_nest_max)
        VRPSX_CONSTANT_METHOD(int, _sc_fsync)
        VRPSX_CONSTANT_METHOD(int, _sc_getgr_r_size_max)
        VRPSX_CONSTANT_METHOD(int, _sc_getpw_r_size_max)
        VRPSX_CONSTANT_METHOD(int, _sc_host_name_max)
        VRPSX_CONSTANT_METHOD(int, _sc_iov_max)
        VRPSX_CONSTANT_METHOD(int, _sc_ipv6)
        VRPSX_CONSTANT_METHOD(int, _sc_job_control)
        VRPSX_CONSTANT_METHOD(int, _sc_line_max)
        VRPSX_CONSTANT_METHOD(int, _sc_login_name_max)
        VRPSX_CONSTANT_METHOD(int, _sc_mapped_files)
        VRPSX_CONSTANT_METHOD(int, _sc_memlock)
        VRPSX_CONSTANT_METHOD(int, _sc_memlock_range)
        VRPSX_CONSTANT_METHOD(int, _sc_memory_protection)
        VRPSX_CONSTANT_METHOD(int, _sc_message_passing)
        VRPSX_CONSTANT_METHOD(int, _sc_monotonic_clock)
        VRPSX_CONSTANT_METHOD(int, _sc_mq_open_max)
        VRPSX_CONSTANT_METHOD(int, _sc_mq_prio_max)
        VRPSX_CONSTANT_METHOD(int, _sc_ngroups_max)
        VRPSX_CONSTANT_METHOD(int, _sc_open_max)
        VRPSX_CONSTANT_METHOD(int, _sc_page_size)
        VRPSX_CONSTANT_METHOD(int, _sc_pagesize)
        VRPSX_CONSTANT_METHOD(int, _sc_prioritized_io)
        VRPSX_CONSTANT_METHOD(int, _sc_priority_scheduling)
        VRPSX_CONSTANT_METHOD(int, _sc_raw_sockets)
        VRPSX_CONSTANT_METHOD(int, _sc_re_dup_max)
        VRPSX_CONSTANT_METHOD(int, _sc_reader_writer_locks)
        VRPSX_CONSTANT_METHOD(int, _sc_realtime_signals)
        VRPSX_CONSTANT_METHOD(int, _sc_regexp)
        VRPSX_CONSTANT_METHOD(int, _sc_rtsig_max)
        VRPSX_CONSTANT_METHOD(int, _sc_saved_ids)
        VRPSX_CONSTANT_METHOD(int, _sc_sem_nsems_max)
        VRPSX_CONSTANT_METHOD(int, _sc_sem_value_max)
        VRPSX_CONSTANT_METHOD(int, _sc_semaphores)
        VRPSX_CONSTANT_METHOD(int, _sc_shared_memory_objects)
        VRPSX_CONSTANT_METHOD(int, _sc_shell)
        VRPSX_CONSTANT_METHOD(int, _sc_sigqueue_max)
        VRPSX_CONSTANT_METHOD(int, _sc_spawn)
        VRPSX_CONSTANT_METHOD(int, _sc_spin_locks)
        VRPSX_CONSTANT_METHOD(int, _sc_sporadic_server)
        VRPSX_CONSTANT_METHOD(int, _sc_ss_repl_max)
        VRPSX_CONSTANT_METHOD(int, _sc_stream_max)
        VRPSX_CONSTANT_METHOD(int, _sc_symloop_max)
        VRPSX_CONSTANT_METHOD(int, _sc_synchronized_io)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_attr_stackaddr)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_attr_stacksize)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_cputime)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_destructor_iterations)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_keys_max)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_prio_inherit)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_prio_protect)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_priority_scheduling)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_process_shared)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_robust_prio_inherit)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_robust_prio_protect)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_safe_functions)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_sporadic_server)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_stack_min)
        VRPSX_CONSTANT_METHOD(int, _sc_thread_threads_max)
        VRPSX_CONSTANT_METHOD(int, _sc_threads)
        VRPSX_CONSTANT_METHOD(int, _sc_timeouts)
        VRPSX_CONSTANT_METHOD(int, _sc_timer_max)
        VRPSX_CONSTANT_METHOD(int, _sc_timers)
        VRPSX_CONSTANT_METHOD(int, _sc_trace)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_event_filter)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_event_name_max)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_inherit)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_log)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_name_max)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_sys_max)
        VRPSX_CONSTANT_METHOD(int, _sc_trace_user_event_max)
        VRPSX_CONSTANT_METHOD(int, _sc_tty_name_max)
        VRPSX_CONSTANT_METHOD(int, _sc_typed_memory_objects)
        VRPSX_CONSTANT_METHOD(int, _sc_tzname_max)
        VRPSX_CONSTANT_METHOD(int, _sc_v6_ilp32_off32)
        VRPSX_CONSTANT_METHOD(int, _sc_v6_ilp32_offbig)
        VRPSX_CONSTANT_METHOD(int, _sc_v6_lp64_off64)
        VRPSX_CONSTANT_METHOD(int, _sc_v6_lpbig_offbig)
        VRPSX_CONSTANT_METHOD(int, _sc_v7_ilp32_off32)
        VRPSX_CONSTANT_METHOD(int, _sc_v7_ilp32_offbig)
        VRPSX_CONSTANT_METHOD(int, _sc_v7_lp64_off64)
        VRPSX_CONSTANT_METHOD(int, _sc_v7_lpbig_offbig)
        VRPSX_CONSTANT_METHOD(int, _sc_version)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_crypt)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_enh_i18n)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_realtime)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_realtime_threads)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_shm)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_streams)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_unix)
        VRPSX_CONSTANT_METHOD(int, _sc_xopen_version)

        VRPSX_CONSTANT_METHOD(int, _pc_2_symlinks)
        VRPSX_CONSTANT_METHOD(int, _pc_alloc_size_min)
        VRPSX_CONSTANT_METHOD(int, _pc_async_io)
        VRPSX_CONSTANT_METHOD(int, _pc_chown_restricted)
        VRPSX_CONSTANT_METHOD(int, _pc_filesizebits)
        VRPSX_CONSTANT_METHOD(int, _pc_link_max)
        VRPSX_CONSTANT_METHOD(int, _pc_max_canon)
        VRPSX_CONSTANT_METHOD(int, _pc_max_input)
        VRPSX_CONSTANT_METHOD(int, _pc_name_max)
        VRPSX_CONSTANT_METHOD(int, _pc_no_trunc)
        VRPSX_CONSTANT_METHOD(int, _pc_path_max)
        VRPSX_CONSTANT_METHOD(int, _pc_pipe_buf)
        VRPSX_CONSTANT_METHOD(int, _pc_prio_io)
        VRPSX_CONSTANT_METHOD(int, _pc_rec_incr_xfer_size)
        VRPSX_CONSTANT_METHOD(int, _pc_rec_max_xfer_size)
        VRPSX_CONSTANT_METHOD(int, _pc_rec_min_xfer_size)
        VRPSX_CONSTANT_METHOD(int, _pc_rec_xfer_align)
        VRPSX_CONSTANT_METHOD(int, _pc_symlink_max)
        VRPSX_CONSTANT_METHOD(int, _pc_sync_io)
        VRPSX_CONSTANT_METHOD(int, _pc_timestamp_resolution)
        VRPSX_CONSTANT_METHOD(int, _pc_vdisable)
    };
}


#endif /* VRPSX_SYSTEM_HPP */
