/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * These are white-box tests of the implementation of vrpsx_system_t
 * defined by the parent directory.
 *
 * By "white-box" we mean that these tests are designed for one specific
 * implementation, and may need adjustment when that implementation changes,
 * even if those changes are backward-compatible at the interface level.
 *
 * In principle one could also define "black-box" tests that require only
 * those behaviors guaranteed by the interface itself, though perhaps with
 * the restriction that they be applied to a particular execution environment.
 * As yet we have not defined such tests, and therefore these white-box tests
 * may overlap what such black-box tests would do.  Feel free to separate out
 * white-box versus black-box testing, if you are of a mind to do so.
 */

#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE   500

#include "test_allocator.h"
#include "test_error.h"
#include "test_timer.h"
#include <vrpsx_system_main.h>
#include <vrpsx_system_main_identifier.h>
#include <vrpsx_identified_function.h>
#include <vrpsx_identified_function_identifier.h>
#include <vrpsx_utility.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <pwd.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#if ! defined(__CYGWIN__)
extern long timezone;
#endif

#if defined(__linux__)

    /* In older Linux environments we must allow GNU extensions
     * generally or else declare strsignal ourselves.  We prefer
     * the latter alternative, at least for now.
     */

    extern char *strsignal(int signo);

#endif


#define MALLOC_AT(WHERE, TYPE)  ((TYPE *)malloc_at((WHERE), sizeof(TYPE)))

#define ZERO(PTR)     (memset((PTR), 0, sizeof *(PTR)))

#define DEAD(PTR)     (memset((PTR), 0xDE, sizeof *(PTR)))


/* The standardization strdup is unclear, so we provide our own, and
 * while we are at it we support a NULL argument by returning NULL.
 */
static char *strdup_or_null(char const *str)
{
    char *copy = NULL;
    if (str) {
        size_t size = strlen(str) + 1;
        copy = (char *)malloc(size);
        if (copy) {
            memcpy(copy, str, size);
        }
    }
    return copy;
}


int test_error_acquire(
    vrpsx_system_t *system)
{
    int result = -1;
    vrpsx_errno_t expected_ern = VRPSX_CONSTANT(system, enotdir);
    vrpsx_system_error_t *error = (vrpsx_system_error_t *)0xBAD;
    VRPSX_CALL2(system, error_acquire,
            expected_ern,
            &error);

    vrpsx_errno_t actual_ern = VRPSX_CALL0(error, number);
    if (actual_ern != expected_ern) {
        PRINT_HERE;
        printf("errno %d != errno %d\n", actual_ern, expected_ern);
        goto fail;
    }

    char const *actual_text = VRPSX_CALL0(error, text);
    char const *expected_text = strerror(expected_ern);
    if (strcmp(actual_text, expected_text)) {
        PRINT_HERE;
        printf("text \"%s\" != text \"%s\"\n", actual_text, expected_text);
        goto fail;
    }

    result = 0;

cleanup:
    VRPSX_COUNTED_REFERENCE_CLEAR(vrpsx_system_error_t, &error);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_memory_allocation(
    vrpsx_system_t *system)
{
    char *block;

    /* Call the operations on the vrpsx_allocator_t superinterface.
     */
    block = (char *)VRPSX_CALL1(&system->allocator, allocate_memory, 1);
    if (! block) {
        PRINT_HERE;
        printf("vrpsx_allocator_t::allocate_memory returned NULL\n");
        return -1;
    }
    VRPSX_CALL1(&system->allocator, free_memory, block);

    /* Call the operations that are defined
     * directly by the vrpsx_system_t interface.
     */
    block = (char *)VRPSX_CALL1(system, allocate_memory, 1);
    if (! block) {
        PRINT_HERE;
        printf("vrpsx_system_t::allocate_memory returned NULL\n");
        return -1;
    }
    VRPSX_CALL1(system, free_memory, block);

    /* The interface comments indicate that corresponding operations
     * on the two interfaces are equivalent; therefore we should be
     * able to mix them.
     */
    block = (char *)VRPSX_CALL1(&system->allocator, allocate_memory, 1);
    if (! block) {
        PRINT_HERE;
        printf("vrpsx_allocator_t::allocate_memory returned NULL\n");
        return -1;
    }
    VRPSX_CALL1(system, free_memory, block);

    block = (char *)VRPSX_CALL1(system, allocate_memory, 1);
    if (! block) {
        PRINT_HERE;
        printf("vrpsx_system_t::allocate_memory returned NULL\n");
        return -1;
    }
    VRPSX_CALL1(&system->allocator, free_memory, block);

    return 0;
}


int test_exit(
    vrpsx_system_t *system)
{
    const int expected_exit_code = 123;
    int actual_status;
    pid_t child;
    int ern;

    child = fork();
    if (child == -1) {
        ern = errno;
        PRINT_HERE;
        printf("fork failed: %s\n", strerror(ern));
        return -1;
    } else if (child == 0) {
        VRPSX_CALL1(system, exit, expected_exit_code);
        fputs("vrpsx_system_t::exit returned\n", stderr);
        exit((expected_exit_code != 3) ? 3 : 5);
    }

    if (waitpid(child, &actual_status, 0) != child) {
        ern = errno;
        PRINT_HERE;
        printf("waitpid failed: %s\n", strerror(ern));
        return -1;
    }

    if (! WIFEXITED(actual_status)) {
        PRINT_HERE;
        printf("child process exited abnormally with status: %#x\n",
                actual_status);
        return -1;
    }

    int actual_exit_code = WEXITSTATUS(actual_status);
    if (actual_exit_code != expected_exit_code) {
        PRINT_HERE;
        printf("unexpected exit code: %d != %d\n",
                actual_exit_code, expected_exit_code);
        return -1;
    }

    return 0;
}


int test_uname(
    vrpsx_system_t *system)
{
    int result = -1;
    struct utsname expected;
    vrpsx_utsname_t *actual = (vrpsx_utsname_t *)0xBAD;
    char const *exp;
    char const *act;
    vrpsx_errno_t ern;

    if (uname(&expected)) {
        ern = errno;
        PRINT_HERE;
        printf("uname failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL1(system, uname, &actual);
    if (ern) {
        printError(HERE, system, "uname", ern);
        return -1;
    }

    exp = expected.machine;
    act = VRPSX_CALL0(actual, machine);
    if (strcmp(act, exp)) {
        PRINT_HERE;
        printf("wrong machine: \"%s\" != \"%s\"\n", act, exp);
        goto fail;
    }

    exp = expected.nodename;
    act = VRPSX_CALL0(actual, nodename);
    if (strcmp(act, exp)) {
        PRINT_HERE;
        printf("wrong nodename: \"%s\" != \"%s\"\n", act, exp);
        goto fail;
    }

    exp = expected.release;
    act = VRPSX_CALL0(actual, release);
    if (strcmp(act, exp)) {
        PRINT_HERE;
        printf("wrong release: \"%s\" != \"%s\"\n", act, exp);
        goto fail;
    }

    exp = expected.sysname;
    act = VRPSX_CALL0(actual, sysname);
    if (strcmp(act, exp)) {
        PRINT_HERE;
        printf("wrong sysname: \"%s\" != \"%s\"\n", act, exp);
        goto fail;
    }

    exp = expected.version;
    act = VRPSX_CALL0(actual, version);
    if (strcmp(act, exp)) {
        PRINT_HERE;
        printf("wrong version: \"%s\" != \"%s\"\n", act, exp);
        goto fail;
    }

    act = VRPSX_CALL0(actual, virtualization);
    if (! act) {
        PRINT_HERE;
        printf("NULL virtualization\n");
        goto fail;
    }
    if (! memchr(act, '\0', 1024)) {
        PRINT_HERE;
        printf("unterminated or very long virtualization\n");
        goto fail;
    }

    result = 0;

cleanup:
    VRPSX_SINGLE_OWNER_CLEAR(vrpsx_utsname_t, &actual);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_sysconf(
    vrpsx_system_t *system)
{
    long expected = 0;
    long actual = 0xBAD;
    vrpsx_errno_t ern;

    errno = 0;
    expected = sysconf(_SC_CLK_TCK);
    if (errno) {
        ern = errno;
        PRINT_HERE;
        printf("sysconf failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL2(system, sysconf,
            VRPSX_CONSTANT(system, _sc_clk_tck),
            &actual);
    if (ern) {
        printError(HERE, system, "sysconf", ern);
        return -1;
    }

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong value: %ld != %ld\n", actual, expected);
        return -1;
    }

    return 0;
}


int test_pathconf(
    vrpsx_system_t *system)
{
    long expected = 0;
    long actual = 0xBAD;
    vrpsx_errno_t ern;

    errno = 0;
    expected = pathconf(".", _PC_NAME_MAX);
    if (errno) {
        ern = errno;
        PRINT_HERE;
        printf("pathconf(\".\", ...) failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL3(system, pathconf,
            ".",
            VRPSX_CONSTANT(system, _pc_name_max),
            &actual);
    if (ern) {
        printError(HERE, system, "pathconf(\".\", ...)", ern);
        return -1;
    }

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong value: %ld != %ld\n", actual, expected);
        return -1;
    }

    return 0;
}


int test_fpathconf(
    vrpsx_system_t *system)
{
    int result = -1;
    int fd = -1;
    long expected = 0;
    long actual = 0xBAD;
    vrpsx_errno_t ern;

    /* If the implementation changes to translate the file descriptors
     * in some way, then this test will need to be modified to open the path
     * in question twice: once directly, and once through vrpsx_system_t.
     */
    fd = open(".", O_RDONLY);
    if (fd == -1) {
        ern = errno;
        PRINT_HERE;
        printf("open(\".\", O_RDONLY) failed: %s\n", strerror(ern));
        return -1;
    }

    errno = 0;
    expected = fpathconf(fd, _PC_NAME_MAX);
    if (errno) {
        ern = errno;
        PRINT_HERE;
        printf("fpathconf failed: %s\n", strerror(ern));
        goto fail;
    }

    ern = VRPSX_CALL3(system, fpathconf,
            fd,
            VRPSX_CONSTANT(system, _pc_name_max),
            &actual);
    if (ern) {
        printError(HERE, system, "fpathconf", ern);
        goto fail;
    }

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong value: %ld != %ld\n", actual, expected);
        goto fail;
    }

    result = 0;

cleanup:
    close(fd);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_getcwd(
    vrpsx_system_t *system)
{
    /* TO DO: Support testing in directories with very long pathnames.
     */
    int result = -1;
    char expected[PATH_MAX];
    test_allocator_t allocator;
    vrpsx_errno_t ern;
    char *actual = (char *)0xBAD;

    test_allocator_construct(&allocator);

    if (! getcwd(expected, sizeof expected)) {
        ern = errno;
        PRINT_HERE;
        printf("getcwd failed: %s\n", strerror(ern));
        goto fail;
    }

    ern = VRPSX_CALL2(system, getcwd,
            &allocator.interface,
            &actual);

    if (strcmp(actual, expected)) {
        PRINT_HERE;
        printf("wrong pathname: \"%s\" != \"%s\"\n", actual, expected);
        goto fail;
    }

    if (assert_singular_allocation(
                HERE, &allocator, actual, strlen(actual) + 1)) {
        goto fail;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto fail;
    }

    result = 0;

cleanup:
    test_allocator_destroy(&allocator);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_get_timezone(
    vrpsx_system_t *system)
{
    long expected;

#if defined(__CYGWIN__)
    expected = _timezone;
#else
    expected = timezone;
#endif

    vrpsx_int64_t actual = VRPSX_CALL0(system, get_timezone);

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong time zone: %lld != %ld\n", (long long)actual, expected);
        return -1;
    }

    return 0;
}


int test_get_tzname(
    vrpsx_system_t *system)
{
    char const *const *actual = VRPSX_CALL0(system, get_tzname);

    if (strcmp(actual[0], tzname[0])) {
        PRINT_HERE;
        printf("wrong standard time zone: \"%s\" != \"%s\"\n",
                actual[0], tzname[0]);
        return -1;
    }

    if (strcmp(actual[1], tzname[1])) {
        PRINT_HERE;
        printf("wrong daylight time zone: \"%s\" != \"%s\"\n",
                actual[1], tzname[1]);
        return -1;
    }

    return 0;
}


#define TM_CMP(FIELD)                                           \
    if (actual.FIELD != expected.FIELD) {                       \
        PRINT_HERE;                                             \
        printf("%ld: wrong %s: %d != %d\n", (long)moment,       \
                #FIELD, actual.FIELD, expected.FIELD);          \
        return -1;                                              \
    }

int test_gmtime(
    vrpsx_system_t *system)
{
    time_t moment;
    vrpsx_tm_t actual;
    struct tm expected;
    vrpsx_errno_t ern;

    if (time(&moment) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    ZERO(&expected);
    if (! gmtime_r(&moment, &expected)) {
        ern = errno;
        PRINT_HERE;
        printf("gmtime_r failed: %s\n", strerror(ern));
        return -1;
    }

    DEAD(&actual);
    ern = VRPSX_CALL2(system, gmtime, moment, &actual);
    if (ern) {
        printError(HERE, system, "gmtime", ern);
        return -1;
    }

    TM_CMP(tm_sec);
    TM_CMP(tm_min);
    TM_CMP(tm_hour);
    TM_CMP(tm_mday);
    TM_CMP(tm_mon);
    TM_CMP(tm_year);
    TM_CMP(tm_wday);
    TM_CMP(tm_yday);
    TM_CMP(tm_isdst);

    return 0;
}

int test_localtime(
    vrpsx_system_t *system)
{
    time_t moment;
    int quarter;
    vrpsx_tm_t actual;
    struct tm expected;
    vrpsx_errno_t ern;

    if (time(&moment) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    /* We test the time now, then 3 more times, adding about one quarter of
     * a year each time, to try to cover both daylight saving time phases.
     */
    for (quarter = 0; quarter < 4; ++quarter, moment += 3*30*24*60*60) {
        ZERO(&expected);
        if (! localtime_r(&moment, &expected)) {
            ern = errno;
            PRINT_HERE;
            printf("localtime_r failed: %s\n", strerror(ern));
            return -1;
        }

        DEAD(&actual);
        ern = VRPSX_CALL2(system, localtime, moment, &actual);
        if (ern) {
            printError(HERE, system, "localtime", ern);
            return -1;
        }

        TM_CMP(tm_sec);
        TM_CMP(tm_min);
        TM_CMP(tm_hour);
        TM_CMP(tm_mday);
        TM_CMP(tm_mon);
        TM_CMP(tm_year);
        TM_CMP(tm_wday);
        TM_CMP(tm_yday);
        TM_CMP(tm_isdst);
    }

    return 0;
}

#undef TM_CMP


int test_mktime(
    vrpsx_system_t *system)
{
    time_t moment;
    struct tm native_input;
    vrpsx_tm_t vrpsx_input;
    time_t expected;
    vrpsx_time_t actual;
    vrpsx_errno_t ern;

    if (time(&moment) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    ZERO(&native_input);
    if (! localtime_r(&moment, &native_input)) {
        ern = errno;
        PRINT_HERE;
        printf("gmtime_r failed: %s\n", strerror(ern));
        return -1;
    }

    DEAD(&vrpsx_input);
    vrpsx_input.tm_sec = native_input.tm_sec;
    vrpsx_input.tm_min = native_input.tm_min;
    vrpsx_input.tm_hour = native_input.tm_hour;
    vrpsx_input.tm_mday = native_input.tm_mday;
    vrpsx_input.tm_mon = native_input.tm_mon;
    vrpsx_input.tm_year = native_input.tm_year;
    vrpsx_input.tm_wday = native_input.tm_wday;
    vrpsx_input.tm_yday = native_input.tm_yday;
    vrpsx_input.tm_isdst = native_input.tm_isdst;

    expected = mktime(&native_input);
    if (expected == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("mktime failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL2(system, mktime, &vrpsx_input, &actual);
    if (ern) {
        printError(HERE, system, "mktime", ern);
        return -1;
    }

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong time: %lld != %lld\n",
                (long long)actual, (long long)expected);
        return -1;
    }

    return 0;
}


int test_time(
    vrpsx_system_t *system)
{
    time_t lo, hi;
    vrpsx_time_t actual;
    vrpsx_errno_t ern;

    if (time(&lo) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL1(system, time, &actual);
    if (ern) {
        printError(HERE, system, "time", ern);
        return -1;
    }

    if (time(&hi) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    if (actual < lo || hi < actual) {
        PRINT_HERE;
        printf("wrong time: %lld not in [%lld, %lld]\n",
                (long long)actual, (long long)lo, (long long)hi);
        return -1;
    }

    return 0;
}


/* Expands to a true expression if the first time
 * is strictly less than the second time.
 *
 * Both macro parameters support both "struct timespec" and "vrpsx_timespec_t",
 * which is the main reason that this is a macro, not a function as such.
 *
 * NOTE: May evaluate the argument expressions multiple times.
 */
#define TIMESPEC_LESS(A, B)                                             \
    ( (A).tv_sec < (B).tv_sec                                           \
      || ( (A).tv_sec == (B).tv_sec && (A).tv_nsec < (B).tv_nsec ) )

int test_clock_gettime(
    vrpsx_system_t *system)
{
#if ! defined(CLOCK_MONOTONIC)

    puts("SKIPPED because this test is unreliable without a monotonic clock");
    return 0;

#else

    struct timespec lo, hi;
    vrpsx_timespec_t actual;
    vrpsx_errno_t ern;

    if (clock_gettime(CLOCK_MONOTONIC, &lo) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    ern = VRPSX_CALL2(system, clock_gettime,
            VRPSX_CONSTANT(system, clock_monotonic),
            &actual);
    if (ern) {
        printError(HERE, system, "clock_gettime", ern);
        return -1;
    }

    if (clock_gettime(CLOCK_MONOTONIC, &hi) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("clock_gettime failed: %s\n", strerror(ern));
        return -1;
    }

    if (TIMESPEC_LESS(actual, lo) || TIMESPEC_LESS(hi, actual)) {
        PRINT_HERE;
        printf("wrong time: %lld.%.9ld not in [%lld.%.9ld, %lld.%.9ld]\n",
                (long long)actual.tv_sec,
                (long)actual.tv_nsec,
                (long long)lo.tv_sec,
                (long)lo.tv_nsec,
                (long long)hi.tv_sec,
                (long)hi.tv_nsec);
        return -1;
    }

    return 0;

#endif
}


int test_nanosleep_uninterrupted(
    vrpsx_system_t *system)
{
    struct timespec lo, hi, actual;
    vrpsx_timespec_t duration, remaining;
    vrpsx_errno_t ern;

    if (clock_gettime(CLOCK_REALTIME, &lo) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        return -1;
    }

    /* 0.15 sec, which we hope is long enough to exceed the time slice
     * granularity even on a virtual machine with coarse time slices,
     * for most test runs.
     */
    duration.tv_sec = 0;
    duration.tv_nsec = 150000000;

    DEAD(&remaining);
    ern = VRPSX_CALL2(system, nanosleep, &duration, &remaining);
    if (ern) {
        printError(HERE, system, "nanosleep", ern);
        return -1;
    }

    if (clock_gettime(CLOCK_REALTIME, &hi) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("clock_gettime failed: %s\n", strerror(ern));
        return -1;
    }

    actual.tv_sec = hi.tv_sec - lo.tv_sec;
    actual.tv_nsec = hi.tv_nsec - lo.tv_nsec;
    if (actual.tv_nsec < 0) {
        actual.tv_nsec += 1000000000;
        actual.tv_sec -= 1;
    }

    if (TIMESPEC_LESS(actual, duration)) {
        PRINT_HERE;
        printf("insufficient duration: %lld.%.9ld < %lld.%.9ld\n",
                (long long)actual.tv_sec,
                (long)actual.tv_nsec,
                (long long)duration.tv_sec,
                (long)duration.tv_nsec);
        return -1;
    }

    if (remaining.tv_sec || remaining.tv_nsec) {
        PRINT_HERE;
        printf("nonzero remaining time: %lld.%.9ld\n",
                (long long)remaining.tv_sec, (long)remaining.tv_nsec);
        return -1;
    }

    return 0;
}


static void do_nothing(int sig) {}

static void *test_nanosleep_interrupted_helper(void *arg)
{
    pthread_t target = *(const pthread_t *)arg;
    struct timespec pause;

    /* 1 microsecond */
    pause.tv_sec = 0;
    pause.tv_nsec = 1000;

    nanosleep(&pause, NULL);

    pthread_kill(target, SIGUSR1);

    return 0;
}

int test_nanosleep_interrupted(
    vrpsx_system_t *system)
{
    void (*old_handler)(int) = NULL;
    pthread_t self, helper, *helper_opt = NULL;
    struct timespec lo, hi, actual;
    vrpsx_timespec_t duration, remaining;
    vrpsx_errno_t ern;

    old_handler = signal(SIGUSR1, &do_nothing);

    if (clock_gettime(CLOCK_REALTIME, &lo) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("time failed: %s\n", strerror(ern));
        goto fail;
    }

    self = pthread_self();
    ern = pthread_create(&helper, NULL,
            &test_nanosleep_interrupted_helper, &self);
    if (ern) {
        PRINT_HERE;
        printf("pthread_create failed: %s\n", strerror(ern));
        goto fail;
    }
    helper_opt = &helper;

    /* 0.1 sec, which we hope is long enough to exceed the time slice
     * granularity even on a virtual machine with coarse time slices,
     * for most test runs.
     */
    duration.tv_sec = 0;
    duration.tv_nsec = 100000000;

    DEAD(&remaining);
    ern = VRPSX_CALL2(system, nanosleep, &duration, &remaining);
    if (ern && ern != VRPSX_CONSTANT(system, eintr)) {
        printError(HERE, system, "nanosleep", ern);
        goto fail;
    }

    if (clock_gettime(CLOCK_REALTIME, &hi) == (time_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("clock_gettime failed: %s\n", strerror(ern));
        goto fail;
    }

    actual.tv_sec = hi.tv_sec - lo.tv_sec;
    actual.tv_nsec = hi.tv_nsec - lo.tv_nsec;
    if (actual.tv_nsec < 0) {
        actual.tv_nsec += 1000000000;
        actual.tv_sec -= 1;
    }

    actual.tv_sec += remaining.tv_sec;
    actual.tv_nsec += remaining.tv_nsec;
    if (actual.tv_nsec >= 1000000000) {
        actual.tv_nsec -= 1000000000;
        actual.tv_sec += 1;
    }

    if (TIMESPEC_LESS(actual, duration)) {
        PRINT_HERE;
        printf("insufficient duration+remaining: %lld.%.9ld < %lld.%.9ld\n",
                (long long)actual.tv_sec,
                (long)actual.tv_nsec,
                (long long)duration.tv_sec,
                (long)duration.tv_nsec);
        goto fail;
    }

    pthread_join(*helper_opt, NULL);
    signal(SIGUSR1, old_handler);
    return 0;

fail:
    if (helper_opt) {
        pthread_kill(*helper_opt, SIGUSR1);
        pthread_join(*helper_opt, NULL);
    }
    signal(SIGUSR1, old_handler);
    return -1;
}


int test_getlogin(
    vrpsx_system_t *system)
{
    int result = -1;
    char *expected = NULL;
    test_allocator_t allocator;
    vrpsx_errno_t ern;
    char *actual = (char *)0xBAD;

    test_allocator_construct(&allocator);

    expected = strdup_or_null(getlogin());
    if (! expected) {
        ern = errno;
        if (ern == ENOTTY) {
            puts("SKIPPED because getlogin failed with ENOTTY");
            result = 0;
            goto cleanup;
        } else if (ern == ENOENT) {
            puts("SKIPPED because getlogin failed with ENOENT");
            result = 0;
            goto cleanup;
        } else {
            PRINT_HERE;
            printf("getlogin failed: %s\n", strerror(ern));
            goto fail;
        }
    }

    ern = VRPSX_CALL2(system, getlogin,
            &allocator.interface,
            &actual);
    if (ern) {
        printError(HERE, system, "getlogin", ern);
        goto fail;
    }

    if (strcmp(actual, expected)) {
        PRINT_HERE;
        printf("wrong login: \"%s\" != \"%s\"\n", actual, expected);
        goto fail;
    }

    if (assert_singular_allocation(
                HERE, &allocator, actual, strlen(actual) + 1)) {
        goto fail;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto fail;
    }

    result = 0;

cleanup:
    test_allocator_destroy(&allocator);
    free(expected);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_getpwnam(
    vrpsx_system_t *system)
{
    int result = -1;
    char *user = NULL;
    vrpsx_passwd_t *actual = (vrpsx_passwd_t *)0xBAD;
    struct passwd *expected;
    vrpsx_errno_t ern;

    user = strdup_or_null(getlogin());
    if (! user) {
        ern = errno;
        if (ern == ENOTTY) {
            puts("SKIPPED because getlogin failed with ENOTTY");
            return 0;
        } else if (ern == ENOENT) {
            puts("SKIPPED because getlogin failed with ENOENT");
            return 0;
        } else {
            PRINT_HERE;
            printf("getlogin failed: %s\n", strerror(ern));
            return -1;
        }
    }

    ern = VRPSX_CALL2(system, getpwnam,
            user,
            &actual);
    if (ern) {
        printError(HERE, system, "getpwnam", ern);
        goto fail;
    }

    if (! (expected = getpwnam(user))) {
        ern = errno;
        PRINT_HERE;
        printf("getpwnam failed: %s\n", strerror(ern));
        goto fail;
    }

#define PW_COMPARE_FIELD_STR(TYPE, FIELD)                               \
    {                                                                   \
        TYPE actual_field = VRPSX_CALL0(actual, FIELD);                 \
        TYPE expected_field = expected -> VRPSX_CAT2(pw_, FIELD);       \
        if (strcmp(actual_field, expected_field)) {                     \
            PRINT_HERE;                                                 \
            printf("wrong " #FIELD ": \"%s\" != \"%s\"\n",              \
                    actual_field, expected_field);                      \
            goto fail;                                                  \
        }                                                               \
    }

#define PW_COMPARE_FIELD_INT(TYPE, FIELD)                               \
    {                                                                   \
        TYPE actual_field = VRPSX_CALL0(actual, FIELD);                 \
        TYPE expected_field = expected -> VRPSX_CAT2(pw_, FIELD);       \
        if (actual_field  != expected_field) {                          \
            PRINT_HERE;                                                 \
            printf("wrong " #FIELD ": %lld != %lld\n",                  \
                    (long long)actual_field, (long long)expected_field); \
            goto fail;                                                  \
        }                                                               \
    }

    PW_COMPARE_FIELD_STR(char const *, dir);
    PW_COMPARE_FIELD_STR(char const *, name);
    PW_COMPARE_FIELD_STR(char const *, shell);
    PW_COMPARE_FIELD_INT(vrpsx_gid_t, gid);
    PW_COMPARE_FIELD_INT(vrpsx_uid_t, uid);

#undef PW_COMPARE_FIELD_STR
#undef PW_COMPARE_FIELD_INT

    result = 0;

cleanup:
    VRPSX_SINGLE_OWNER_CLEAR(vrpsx_passwd_t, &actual);
    free(user);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_getpid(
    vrpsx_system_t *system)
{
    vrpsx_pid_t actual = VRPSX_CALL0(system, getpid);
    pid_t expected = getpid();

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong pid: %lld != %lld\n",
                (long long)actual, (long long)expected);
        return -1;
    }

    return 0;
}


int test_getpgrp(
    vrpsx_system_t *system)
{
    vrpsx_pid_t actual = VRPSX_CALL0(system, getpgrp);
    pid_t expected = getpgrp();

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong process group: %lld != %lld\n",
                (long long)actual, (long long)expected);
        return -1;
    }

    return 0;
}


int test_getpgid(
    vrpsx_system_t *system)
{
    vrpsx_pid_t pid = getpid();
    vrpsx_pid_t actual;
    pid_t expected;
    vrpsx_errno_t ern;

    ern = VRPSX_CALL2(system, getpgid,
            pid,
            &actual);
    if (ern) {
        printError(HERE, system, "getpgid", ern);
        return -1;
    }

    expected = getpgid(pid);
    if (expected == (pid_t)-1) {
        ern = errno;
        PRINT_HERE;
        printf("getpgid failed: %s\n", strerror(ern));
        return -1;
    }

    if (actual != expected) {
        PRINT_HERE;
        printf("wrong process group: %lld != %lld\n",
                (long long)actual, (long long)expected);
        return -1;
    }

    return 0;
}


int test_kill(
    vrpsx_system_t *system)
{
    int actual_status;
    int actual_signal;
    pid_t child;
    vrpsx_errno_t ern;

    child = fork();
    if (child == -1) {
        ern = errno;
        PRINT_HERE;
        printf("fork failed: %s\n", strerror(ern));
        return -1;
    } else if (child == 0) {
        sleep(100);
        exit(1);
    }

    ern = VRPSX_CALL2(system, kill,
            child,
            VRPSX_CONSTANT(system, sigterm));
    if (ern) {
        printError(HERE, system, "kill", ern);
        waitpid(child, &actual_status, 0);
        return -1;
    }

    if (waitpid(child, &actual_status, 0) != child) {
        ern = errno;
        PRINT_HERE;
        printf("waitpid failed: %s\n", strerror(ern));
        return -1;
    }

    if (! WIFSIGNALED(actual_status)) {
        PRINT_HERE;
        printf("child process not killed: status: %#x\n",
                (unsigned)actual_status);
        return -1;
    }

    actual_signal = WTERMSIG(actual_status);
    if (actual_signal != SIGTERM) {
        PRINT_HERE;
        printf("child process killed by wrong signal: %d (\"%s\")\n",
                actual_signal, strsignal(actual_signal));
        return -1;
    }

    return 0;
}


static int volatile test_raise_signo = !SIGUSR1;

static void test_raise_handler(
    int signo)
{
    test_raise_signo = signo;
}

int test_raise(
    vrpsx_system_t *system)
{
    int result = -1;
    int const expected = SIGUSR1;
    int actual;
    void (*old)(int) = 0;
    vrpsx_errno_t ern;

    test_raise_signo = !SIGUSR1;
    old = signal(expected, &test_raise_handler);

    ern = VRPSX_CALL1(system, raise,
            VRPSX_CONSTANT(system, sigusr1));
    if (ern) {
        printError(HERE, system, "raise", ern);
        goto fail;
    }

    /* POSIX "raise" is not supposed to return before the signal handler
     * that we installed returns.  The definition of vrpsx_system_t does
     * not exempt its "raise" method from this requirement.  Therefore at
     * this point we expect the signal handler to have set test_raise_signal.
     */

    actual = test_raise_signo;
    if (actual != expected) {
        PRINT_HERE;
        printf("wrong signal number: %d (\"%s\")\n",
                actual, strsignal(actual));
        goto fail;
    }

    result = 0;

cleanup:
    signal(expected, old);
    test_raise_signo = !SIGUSR1;
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_strsignal(
    vrpsx_system_t *system)
{
    int result = -1;
    char *actual = strdup_or_null(VRPSX_CALL1(system, strsignal,
                    VRPSX_CONSTANT(system, sigint)));
    char *expected = strdup_or_null(strsignal(SIGINT));

    if (! actual) {
        PRINT_HERE;
        printf("NULL text for vrpsx_system_t::sigint\n");
        goto fail;
    }

    if (! expected) {
        PRINT_HERE;
        printf("NULL text for SIGINT\n");
        goto fail;
    }

    if (strcmp(actual, expected)) {
        PRINT_HERE;
        printf("wrong text for SIGINT: \"%s\" != \"%s\"\n", actual, expected);
        goto fail;
    }

    if (! VRPSX_CALL1(system, strsignal, 10000)) {
        PRINT_HERE;
        printf("NULL text for 10000\n");
        goto fail;
    }

    result = 0;

cleanup:
    free(expected);
    free(actual);
    return result;

fail:
    result = -1;
    goto cleanup;
}


/* Assert that every byte in [begin, end) still has value 0xDE,
 * which is our convention for "dead" memory: memory that has
 * not been written by the code under test.  (Of course, there
 * is a small chance that the code under test will write 0xDE,
 * which we would not detect; that is a natural limitation.)
 */
static int assert_dead(
    where_t where,
    char const *begin,
    char const *end,
    int after)
{
    for (; begin != end; ++begin) {
        if ((unsigned char)*begin != 0xDE) {
            print_where(where);
            printf("buffer %s\n", (after ? "overrun" : "underrun"));
            return -1;
        }
    }

    return 0;
}

static int sigset_init(
    where_t where,
    vrpsx_system_t *system,
    void *space,
    vrpsx_sigset_t **set)
{
    vrpsx_errno_t ern;

    *set = NULL;

    ern = VRPSX_CALL2(system, sigset_init, space, set);
    if (ern) {
        printError(where, system, "sigset_init", ern);
        return -1;
    }

    return 0;
}

static int sigset_empty(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t *set)
{
    vrpsx_errno_t ern = VRPSX_CALL0(set, empty);
    if (ern) {
        printError(where, system, "vrpsx_sigset_t::empty", ern);
        return -1;
    }
    return 0;
}

static int sigset_fill(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t *set)
{
    vrpsx_errno_t ern = VRPSX_CALL0(set, fill);
    if (ern) {
        printError(where, system, "vrpsx_sigset_t::fill", ern);
        return -1;
    }
    return 0;
}

static int sigset_del(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t *set,
    vrpsx_signo_t signo)
{
    vrpsx_errno_t ern = VRPSX_CALL1(set, del, signo);
    if (ern) {
        printError(where, system, "vrpsx_sigset_t::del", ern);
        return -1;
    }
    return 0;
}

static int sigset_add(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t *set,
    vrpsx_signo_t signo)
{
    vrpsx_errno_t ern = VRPSX_CALL1(set, add, signo);
    if (ern) {
        printError(where, system, "vrpsx_sigset_t::add", ern);
        return -1;
    }
    return 0;
}

static int sigset_ismember(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t const *set,
    vrpsx_signo_t signo,
    int *actual)
{
    vrpsx_errno_t ern = VRPSX_CALL2(set, ismember, signo, actual);
    if (ern) {
        static char const fmt[] = "sigset::ismember(%d)";
        char buf[sizeof fmt + 1 + (sizeof signo * CHAR_BIT + 2) / 3 + 1];
        sprintf(buf, fmt, signo);
        printError(where, system, buf, ern);
        return -1;
    }
    return 0;
}

static int assert_sigset_membership(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_sigset_t const *set,
    vrpsx_signo_t signo,
    int expected)
{
    int actual, result;

    result = sigset_ismember(where, system, set, signo, &actual);
    if (result) {
        return result;
    }

    if (actual != expected) {
        print_where(where);
        printf("signal set membership for %d (\"%s\"): %d != %d\n",
                signo, strsignal(signo), actual, expected);
        return -1;
    }

    return 0;
}

static vrpsx_signo_t signo_constant(
    vrpsx_system_t *system,
    size_t offset)
{
    vrpsx_system_vtbl_t const *vtbl = system->vtbl;
    vrpsx_signo_t const *signo
            = (vrpsx_signo_t const *)((char *)vtbl + offset);
    return *signo;
}

static size_t const signo_offsets[] = {
    offsetof(vrpsx_system_vtbl_t, sigabrt),
    offsetof(vrpsx_system_vtbl_t, sigalrm),
    offsetof(vrpsx_system_vtbl_t, sigbus),
    offsetof(vrpsx_system_vtbl_t, sigchld),
    offsetof(vrpsx_system_vtbl_t, sigcont),
    offsetof(vrpsx_system_vtbl_t, sigfpe),
    offsetof(vrpsx_system_vtbl_t, sighup),
    offsetof(vrpsx_system_vtbl_t, sigill),
    offsetof(vrpsx_system_vtbl_t, sigint),
    offsetof(vrpsx_system_vtbl_t, sigkill),
    offsetof(vrpsx_system_vtbl_t, sigpipe),
    offsetof(vrpsx_system_vtbl_t, sigpoll),
    offsetof(vrpsx_system_vtbl_t, sigprof),
    offsetof(vrpsx_system_vtbl_t, sigquit),
    offsetof(vrpsx_system_vtbl_t, sigsegv),
    offsetof(vrpsx_system_vtbl_t, sigstop),
    offsetof(vrpsx_system_vtbl_t, sigsys),
    offsetof(vrpsx_system_vtbl_t, sigterm),
    offsetof(vrpsx_system_vtbl_t, sigtrap),
    offsetof(vrpsx_system_vtbl_t, sigtstp),
    offsetof(vrpsx_system_vtbl_t, sigttin),
    offsetof(vrpsx_system_vtbl_t, sigttou),
    offsetof(vrpsx_system_vtbl_t, sigurg),
    offsetof(vrpsx_system_vtbl_t, sigusr1),
    offsetof(vrpsx_system_vtbl_t, sigusr2),
    offsetof(vrpsx_system_vtbl_t, sigvtalrm),
    offsetof(vrpsx_system_vtbl_t, sigxcpu),
    offsetof(vrpsx_system_vtbl_t, sigxfsz),
    0
};

int test_sigset(
    vrpsx_system_t *system)
{
    int result = -1;
    char *space = NULL, *pad;
    char *space2 = NULL, *pad2;
    vrpsx_sigset_t *set = (vrpsx_sigset_t *)0xBAD;
    vrpsx_sigset_t *set2 = (vrpsx_sigset_t *)0xBAD;
    size_t align = VRPSX_CONSTANT(system, alignof_sigset);
    size_t size = VRPSX_CONSTANT(system, sizeof_sigset);
    size_t pad_size = 16;
    size_t const *signo_offset;

    /* The pointer returned by malloc() will have alignment suitable
     * for any data type.  Therefore by skipping the first "align"
     * bytes of that returned block, we ensure that the alignment of
     * the remainder of the block is no greater than "align", so that
     * if the code under test actually requires a greater alignment
     * without declaring it, we have some hope of noticing the bug.
     *
     * Also, we add some padding at the end to check for buffer overrun.
     */
    space = (char *)malloc_at(HERE, align + size + pad_size);
    if (! space) {
        goto fail;
    }
    memset(space, 0xDE, align + size + pad_size);
    pad = space + align + size;

    if (sigset_init(HERE, system, space + align, &set)) {
        goto fail;
    }

    if (assert_dead(HERE, space, space + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad, pad + pad_size, 1)) {
        goto fail;
    }

    /* Initially, the set should contain no signal numbers.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, set, signo, 0)) {
            goto fail;
        }
    }

    if (sigset_fill(HERE, system, set)) {
        goto fail;
    }

    /* After filling the set, it should contain all signal numbers.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, set, signo, 1)) {
            goto fail;
        }
    }

    if (sigset_del(HERE, system, set, VRPSX_CONSTANT(system, sigint))) {
        goto fail;
    }

    /* After removing SIGINT, it should be a non-member.
     */
    if (assert_sigset_membership(
                HERE, system, set, VRPSX_CONSTANT(system, sigint), 0)) {
        goto fail;
    }

    /* Make a second signal set object.
     */
    space2 = (char *)malloc_at(HERE, align + size + pad_size);
    if (! space2) {
        goto fail;
    }
    memset(space2, 0xDE, align + size + pad_size);
    pad2 = space2 + align + size;

    if (sigset_init(HERE, system, space2 + align, &set2)) {
        goto fail;
    }

    /* Use assign_sigset to make it into a copy of the first signal set.
     */
    VRPSX_CALL1(set2, assign_sigset, set);

    if (sigset_add(HERE, system, set, VRPSX_CONSTANT(system, sigint))) {
        goto fail;
    }

    /* After re-adding SIGINT, it should be a member once again.
     */
    if (assert_sigset_membership(
                HERE, system, set, VRPSX_CONSTANT(system, sigint), 1)) {
        goto fail;
    }

    /* The copy we made before re-adding SIGINT should still omit SIGINT
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(
                    HERE, system, set2, signo,
                    signo != VRPSX_CONSTANT(system, sigint))) {
            goto fail;
        }
    }

    /* Remove SIGABRT from the original set.
     */
    if (sigset_del(HERE, system, set, VRPSX_CONSTANT(system, sigabrt))) {
        goto fail;
    }

    VRPSX_CALL1(set2, swap_sigset, set);

    /* Check the result of the swap:
     *
     *   - set should contain all signals except SIGINT.
     *
     *   - set2 should contain all signals except SIGABRT.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(
                    HERE, system, set, signo,
                    signo != VRPSX_CONSTANT(system, sigint))) {
            goto fail;
        }
    }
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(
                    HERE, system, set2, signo,
                    signo != VRPSX_CONSTANT(system, sigabrt))) {
            goto fail;
        }
    }

    if (sigset_empty(HERE, system, set)) {
        goto fail;
    }

    /* After emptying the set, it should contain no signal numbers.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, set, signo, 0)) {
            goto fail;
        }
    }

    if (assert_dead(HERE, space, space + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad, pad + pad_size, 1)) {
        goto fail;
    }

    if (assert_dead(HERE, space2, space2 + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad2, pad2 + pad_size, 1)) {
        goto fail;
    }

    result = 0;

cleanup:
    free(space2);
    free(space);
    return result;

fail:
    result = -1;
    goto cleanup;
}


static vrpsx_sa_flags_t sa_flags_constant(
    vrpsx_system_t *system,
    size_t offset)
{
    vrpsx_system_vtbl_t const *vtbl = system->vtbl;
    vrpsx_sa_flags_t const *flag
            = (vrpsx_sa_flags_t const *)((char *)vtbl + offset);
    return *flag;
}

typedef struct test_sa_flags_symbol_struct {
    size_t offset;
    char const *name;
} test_sa_flags_symbol_t;

static test_sa_flags_symbol_t const sa_flags_symbols[] = {
    { offsetof(vrpsx_system_vtbl_t, sa_nocldstop), "sa_nocldstop" },
    { offsetof(vrpsx_system_vtbl_t, sa_nocldwait), "sa_nocldwait" },
    { offsetof(vrpsx_system_vtbl_t, sa_nodefer), "sa_nodefer" },
    { offsetof(vrpsx_system_vtbl_t, sa_onstack), "sa_onstack" },
    { offsetof(vrpsx_system_vtbl_t, sa_resethand), "sa_resethand" },
    { offsetof(vrpsx_system_vtbl_t, sa_restart), "sa_restart" },
    { offsetof(vrpsx_system_vtbl_t, sa_siginfo), "sa_siginfo" },
    { 0, NULL }
};

static void print_sa_flags(
    FILE *fp,
    vrpsx_system_t *system,
    vrpsx_sa_flags_t flags)
{
    char const *separator = " ";
    test_sa_flags_symbol_t const *symbol;
    vrpsx_sa_flags_t mask;

    fputs("sa_flags {", fp);

    for (symbol = sa_flags_symbols; symbol->name; ++symbol) {
        mask = sa_flags_constant(system, symbol->offset);
        if ((flags & mask) == mask) {
            fputs(separator, fp);
            fputs(symbol->name, fp);
            separator = " | ";
        }
    }

    fputs(" }", fp);
}

static int assert_sigaction_sa_flags(
    where_t where,
    vrpsx_system_t *system,
    char const *prefix,
    vrpsx_sigaction_t const *action,
    vrpsx_sa_flags_t expected)
{
    vrpsx_sa_flags_t actual = VRPSX_CALL0(action, get_flags);
    if (actual != expected) {
        print_where(where);
        printf("%s: ", prefix);
        print_sa_flags(stdout, system, actual);
        fputs(" != ", stdout);
        print_sa_flags(stdout, system, expected);
        fputc('\n', stdout);
        return -1;
    }

    return 0;
}

typedef enum test_sa_callback_e {
    TEST_SA_UNKNOWN,
    TEST_SA_HANDLER,
    TEST_SA_SIGACTION
} test_sa_callback_t;

static test_sa_callback_t volatile test_sa_callback = TEST_SA_UNKNOWN;

static int volatile test_sa_signal = 0;

static void test_sa_handler(
    int signal_number)
{
    test_sa_callback = TEST_SA_HANDLER;
    test_sa_signal = signal_number;
}

static void test_sa_sigaction(
    int signal_number,
    vrpsx_siginfo_t *information,
    void *context)
{
    test_sa_callback = TEST_SA_SIGACTION;
    test_sa_signal = signal_number;
}

int test_sigaction(
    vrpsx_system_t *system)
{
    int result = -1;
    char *space = NULL, *pad;
    char *space2 = NULL, *pad2;
    char *space3 = NULL, *pad3;
    vrpsx_sigaction_t *action = (vrpsx_sigaction_t *)0xBAD;
    vrpsx_sigaction_t *action2 = (vrpsx_sigaction_t *)0xBAD;
    vrpsx_sigaction_t *action3 = (vrpsx_sigaction_t *)0xBAD;
    vrpsx_errno_t ern;
    size_t align = VRPSX_CONSTANT(system, alignof_sigaction);
    size_t size = VRPSX_CONSTANT(system, sizeof_sigaction);
    size_t pad_size = 16;
    vrpsx_sa_flags_t actual_flags;
    vrpsx_sa_handler_t *actual_handler;
    vrpsx_sa_sigaction_t *actual_sigaction;
    vrpsx_sigset_t const *mask_const;
    vrpsx_sigset_t *mask;
    size_t const *signo_offset;

    /* The pointer returned by malloc() will have alignment suitable
     * for any data type.  Therefore by skipping the first "align"
     * bytes of that returned block, we ensure that the alignment of
     * the remainder of the block is no greater than "align", so that
     * if the code under test actually requires a greater alignment
     * without declaring it, we have some hope of noticing the bug.
     *
     * Also, we add some padding at the end to check for buffer overrun.
     */
    space = (char *)malloc_at(HERE, align + size + pad_size);
    if (! space) {
        goto fail;
    }
    memset(space, 0xDE, align + size + pad_size);
    pad = space + align + size;

    ern = VRPSX_CALL2(system, sigaction_init, space + align, &action);
    if (ern) {
        printError(HERE, system, "sigaction_init",ern);
        goto fail;
    }

    if (assert_dead(HERE, space, space + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad, pad + pad_size, 1)) {
        goto fail;
    }

    if (assert_sigaction_sa_flags(HERE, system,
                    "initially, flags should be clear", action, 0)) {
        goto fail;
    }

    /* Initially, the handlers should be NULL.
     */
    actual_handler = VRPSX_CALL0(action, get_handler);
    if (actual_handler != 0) {
        PRINT_HERE;
        printf("initially, handler should be NULL: %p != NULL\n",
                (const void *)actual_handler);
        goto fail;
    }
    actual_sigaction = VRPSX_CALL0(action, get_sigaction);
    if (actual_sigaction != 0) {
        PRINT_HERE;
        printf("initially, sigaction should be NULL: %p != NULL\n",
                (const void *)actual_sigaction);
        goto fail;
    }

    /* Initially, the signal set should be empty.
     */
    mask_const = VRPSX_CALL0(action, get_mask_const);
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, mask_const, signo, 0)) {
            goto fail;
        }
    }

    /* Modify the mask of the action.
     */
    mask = VRPSX_CALL0(action, get_mask);
    ern = VRPSX_CALL0(mask, fill);
    if (ern) {
        printError(HERE, system, "sigset::fill", ern);
        goto fail;
    }

    /* After filling the action mask, it should contain all signal numbers.
     *
     * Note that we filled through the modifiable interface, but we are reading
     * the results through the non-modifiable interface.  That should work.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, mask_const, signo, 1)) {
            goto fail;
        }
    }

    /* Check effects of set_sigaction.
     */
    VRPSX_CALL1(action, set_sigaction, &test_sa_sigaction);

    if (assert_sigaction_sa_flags(HERE, system,
                    "set_sigaction should set sa_siginfo", action,
                    VRPSX_CONSTANT(system, sa_siginfo))) {
        goto fail;
    }

    actual_handler = VRPSX_CALL0(action, get_handler);
    if (actual_handler != 0) {
        PRINT_HERE;
        printf("legacy handler should still be NULL: %p != NULL\n",
                (const void *)actual_handler);
        goto fail;
    }
    actual_sigaction = VRPSX_CALL0(action, get_sigaction);
    if (actual_sigaction != &test_sa_sigaction) {
        PRINT_HERE;
        printf("sigaction should have been changed: %p != %p\n",
                (const void *)actual_sigaction,
                (const void *)&test_sa_sigaction);
        goto fail;
    }

    /* Check effects of set_flags.
     */
    VRPSX_CALL1(action, set_flags, VRPSX_CONSTANT(system, sa_resethand));

    if (assert_sigaction_sa_flags(HERE, system,
                    "set_flags should set supported flags except sa_siginfo",
                    action, (VRPSX_CONSTANT(system, sa_siginfo)
                            | VRPSX_CONSTANT(system, sa_resethand)))) {
        goto fail;
    }

    /* Check effects of set_flags.
     */
    VRPSX_CALL1(action, set_flags, VRPSX_CONSTANT(system, sa_restart));

    if (assert_sigaction_sa_flags(HERE, system,
                    "set_flags should set supported flags except sa_siginfo",
                    action, (VRPSX_CONSTANT(system, sa_siginfo)
                            | VRPSX_CONSTANT(system, sa_restart)))) {
        goto fail;
    }

    /* Check effects of set_handler.
     */
    VRPSX_CALL1(action, set_handler, &test_sa_handler);

    if (assert_sigaction_sa_flags(HERE, system,
                    "set_handler should clear sa_siginfo",
                    action, VRPSX_CONSTANT(system, sa_restart))) {
        goto fail;
    }

    actual_handler = VRPSX_CALL0(action, get_handler);
    if (actual_handler != &test_sa_handler) {
        PRINT_HERE;
        printf("legacy handler should have been set: %p != %p\n",
                (const void *)actual_handler, (const void *)&test_sa_handler);
        goto fail;
    }
    actual_sigaction = VRPSX_CALL0(action, get_sigaction);
    if (actual_sigaction != 0) {
        PRINT_HERE;
        printf("sigaction should have been cleared: %p != NULL\n",
                (const void *)actual_sigaction);
        goto fail;
    }

    /* Create a second action.
     */
    space2 = (char *)malloc_at(HERE, align + size + pad_size);
    if (! space2) {
        goto fail;
    }
    memset(space2, 0xDE, align + size + pad_size);
    pad2 = space2 + align + size;

    ern = VRPSX_CALL2(system, sigaction_init, space2 + align, &action2);
    if (ern) {
        printError(HERE, system, "sigaction_init", ern);
        goto fail;
    }

    /* Use assign_sigaction to make it into a copy of the first action.
     */
    VRPSX_CALL1(action2, assign_sigaction, action);

    /* The flags should have been copied.
     */
    if (assert_sigaction_sa_flags(HERE, system,
                    "flags should have been copied to new action instance",
                    action2, VRPSX_CONSTANT(system, sa_restart))) {
        goto fail;
    }

    /* The signal mask should have been copied.
     */
    mask_const = VRPSX_CALL0(action2, get_mask_const);
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, mask_const, signo, 1)) {
            goto fail;
        }
    }

    /* Changing the original flags should not affect the copied flags.
     */
    VRPSX_CALL1(action, set_flags, VRPSX_CONSTANT(system, sa_resethand));
    if (assert_sigaction_sa_flags(HERE, system,
                    "copied flags should not have changed",
                    action2, VRPSX_CONSTANT(system, sa_restart))) {
        goto fail;
    }

    /* Changing the original mask should not affect the copied mask.
     */
    mask = VRPSX_CALL0(action, get_mask);
    ern = VRPSX_CALL1(mask, del, VRPSX_CONSTANT(system, sigint));
    if (ern) {
        printError(HERE, system, "sigaction::del", ern);
        goto fail;
    }
    mask_const = VRPSX_CALL0(action2, get_mask_const);
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, mask_const, signo, 1)) {
            goto fail;
        }
    }

    mask = VRPSX_CALL0(action2, get_mask);
    ern = VRPSX_CALL1(mask, del, VRPSX_CONSTANT(system, sigabrt));
    VRPSX_CALL1(action2, set_sigaction, &test_sa_sigaction);

    /* Check the result of the swap.
     */
    VRPSX_CALL1(action2, swap_sigaction, action);

    if (assert_sigaction_sa_flags(HERE, system,
                    "flags incorrectly swapped",
                    action, (VRPSX_CONSTANT(system, sa_siginfo)
                            | VRPSX_CONSTANT(system, sa_restart)))) {
        goto fail;
    }
    if (assert_sigaction_sa_flags(HERE, system,
                    "flags incorrectly swapped",
                    action2, VRPSX_CONSTANT(system, sa_resethand))) {
        goto fail;
    }
    actual_sigaction = VRPSX_CALL0(action, get_sigaction);
    if (actual_sigaction != &test_sa_sigaction) {
        PRINT_HERE;
        printf("sigaction incorrectly swapped: %p != %p\n",
                (const void *)actual_sigaction,
                (const void *)&test_sa_sigaction);
        goto fail;
    }
    actual_handler = VRPSX_CALL0(action2, get_handler);
    if (actual_handler != &test_sa_handler) {
        PRINT_HERE;
        printf("handler incorrectly swapped: %p != %p\n",
                (const void *)actual_handler, (const void *)&test_sa_handler);
        goto fail;
    }
    mask_const = VRPSX_CALL0(action, get_mask_const);
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(HERE, system, mask_const, signo,
                        signo != VRPSX_CONSTANT(system, sigabrt))) {
            goto fail;
        }
    }
    mask_const = VRPSX_CALL0(action2, get_mask_const);
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);
        if (assert_sigset_membership(
                    HERE, system, mask_const, signo,
                    signo != VRPSX_CONSTANT(system, sigint))) {
            goto fail;
        }
    }

    /* Create a third action to receive prior state.
     */
    space3 = (char *)malloc_at(HERE, align + size + pad_size);
    if (! space3) {
        goto fail;
    }
    memset(space3, 0xDE, align + size + pad_size);
    pad3 = space3 + align + size;

    ern = VRPSX_CALL2(system, sigaction_init, space3 + align, &action3);
    if (ern) {
        printError(HERE, system, "sigaction_init",ern);
        goto fail;
    }

    /* Actually install a signal action (as opposed to a handler);
     * check that the prior action is the default.
     */
    ern = VRPSX_CALL3(system, sigaction,
            VRPSX_CONSTANT(system, sigusr1),
            action,
            action3);
    if (ern) {
        printError(HERE, system, "sigaction", ern);
        goto fail;
    }

    actual_flags = VRPSX_CALL0(action3, get_flags);
    if (actual_flags & VRPSX_CONSTANT(system, sa_siginfo)) {
        PRINT_HERE;
        printf("prior state shows an action, not a handler\n");
        goto fail;
    }

    actual_handler = VRPSX_CALL0(action3, get_handler);
    if (actual_handler != VRPSX_CONSTANT(system, sig_dfl)) {
        PRINT_HERE;
        printf("prior state shows non-default handler: %p != %p\n",
                (const void *)actual_handler, (const void *)&test_sa_handler);
        goto fail;
    }

    /* Actually install a signal handler (as opposed to an action).
     */
    ern = VRPSX_CALL3(system, sigaction,
            VRPSX_CONSTANT(system, sigusr2),
            action2,
            NULL);
    if (ern) {
        printError(HERE, system, "sigaction", ern);
        goto fail;
    }

    /* Raise signal to trigger action and check effects.
     */
    test_sa_callback = TEST_SA_UNKNOWN;
    test_sa_signal = 0;

    ern = VRPSX_CALL1(system, raise, VRPSX_CONSTANT(system, sigusr1));
    if (ern) {
        printError(HERE, system, "raise", ern);
        goto fail;
    }

    if (test_sa_callback == TEST_SA_UNKNOWN) {
        PRINT_HERE;
        printf("did not install action or handler\n");
        goto fail;
    }
    if (test_sa_callback != TEST_SA_SIGACTION) {
        PRINT_HERE;
        printf("did not install sa_sigaction\n");
        goto fail;
    }
    if (test_sa_signal != SIGUSR1) {
        PRINT_HERE;
        printf("wrong signal number: %d instead of %d\n",
                test_sa_signal, SIGUSR1);
        goto fail;
    }

    /* Raise signal to trigger handler and check effects.
     */
    test_sa_callback = TEST_SA_UNKNOWN;
    test_sa_signal = 0;

    ern = VRPSX_CALL1(system, raise, VRPSX_CONSTANT(system, sigusr2));
    if (ern) {
        printError(HERE, system, "raise", ern);
        goto fail;
    }

    if (test_sa_callback == TEST_SA_UNKNOWN) {
        PRINT_HERE;
        printf("did not install action or handler\n");
        goto fail;
    }
    if (test_sa_callback != TEST_SA_HANDLER) {
        PRINT_HERE;
        printf("did not install handler\n");
        goto fail;
    }
    if (test_sa_signal != SIGUSR2) {
        PRINT_HERE;
        printf("wrong signal number: %d instead of %d\n",
                test_sa_signal, SIGUSR2);
        goto fail;
    }

    /* Reinstall a signal action; check the the prior action.
     */
    ern = VRPSX_CALL3(system, sigaction,
            VRPSX_CONSTANT(system, sigusr1),
            action,
            action3);
    if (ern) {
        printError(HERE, system, "sigaction", ern);
        goto fail;
    }

    actual_flags = VRPSX_CALL0(action3, get_flags);
    if (! (actual_flags & VRPSX_CONSTANT(system, sa_siginfo))) {
        PRINT_HERE;
        printf("prior state shows a handler, not an action\n");
        goto fail;
    }

    actual_sigaction = VRPSX_CALL0(action3, get_sigaction);
    if (actual_sigaction != &test_sa_sigaction) {
        PRINT_HERE;
        printf("prior sigaction incorrect: %p != %p\n",
                (const void *)actual_sigaction,
                (const void *)&test_sa_sigaction);
        goto fail;
    }

    /* Check that our various API calls did not stray outside
     * the bounds of their various memory buffer parameters.
     */
    if (assert_dead(HERE, space, space + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad, pad + pad_size, 1)) {
        goto fail;
    }

    if (assert_dead(HERE, space2, space2 + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad2, pad2 + pad_size, 1)) {
        goto fail;
    }

    if (assert_dead(HERE, space3, space3 + align, 0)) {
        goto fail;
    }

    if (assert_dead(HERE, pad3, pad3 + pad_size, 1)) {
        goto fail;
    }

    result = 0;

cleanup:
    signal(SIGUSR2, SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    free(space3);
    free(space2);
    free(space);
    return result;

fail:
    result = -1;
    goto cleanup;
}


static int mutex_create(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_mutex_t **mutex)
{
    vrpsx_errno_t ern = VRPSX_CALL1(system, mutex_create, mutex);
    if (ern) {
        printError(where, system, "mutex_create", ern);
        return -1;
    }
    return 0;
}

static int mutex_disown(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_mutex_t *mutex)
{
    vrpsx_errno_t ern;
    VRPSX_SINGLE_OWNER_CLOSE_SROE(vrpsx_mutex_t, &mutex, &ern, NULL);
    if (ern) {
        printError(where, system, "vrpsx_mutex_t::disown", ern);
        return -1;
    }
    return 0;
}

static int mutex_lock(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_mutex_t *mutex)
{
    vrpsx_errno_t ern = VRPSX_CALL0(mutex, lock);
    if (ern) {
        printError(where, system, "vrpsx_mutex_t::lock", ern);
        return -1;
    }
    return 0;
}

static int mutex_unlock(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_mutex_t *mutex)
{
    vrpsx_errno_t ern = VRPSX_CALL0(mutex, unlock);
    if (ern) {
        printError(where, system, "vrpsx_mutex_t::unlock", ern);
        return -1;
    }
    return 0;
}


static int cond_create(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_cond_t **cond)
{
    vrpsx_errno_t ern = VRPSX_CALL1(system, cond_create, cond);
    if (ern) {
        printError(where, system, "cond_create", ern);
        return -1;
    }
    return 0;
}

static int cond_disown(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_cond_t *cond)
{
    vrpsx_errno_t ern;
    VRPSX_SINGLE_OWNER_CLOSE_SROE(vrpsx_cond_t, &cond, &ern, NULL);
    if (ern) {
        printError(where, system, "vrpsx_cond_t::disown", ern);
        return -1;
    }
    return 0;
}

static int cond_broadcast(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_cond_t *cond)
{
    vrpsx_errno_t ern = VRPSX_CALL0(cond, broadcast);
    if (ern) {
        printError(where, system, "vrpsx_cond_t::broadcast", ern);
        return -1;
    }
    return 0;
}

static int cond_wait(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_cond_t *cond,
    vrpsx_mutex_t *mutex)
{
    vrpsx_errno_t ern = VRPSX_CALL1(cond, wait, mutex);
    if (ern) {
        printError(where, system, "vrpsx_cond_t::wait", ern);
        return -1;
    }
    return 0;
}


typedef enum sigwait_step_enum {
    SIGWAIT_STEP_PREPARE,
    SIGWAIT_STEP_IMMINENT,
    SIGWAIT_STEP_DONE
} sigwait_step_t;

typedef struct sigwait_state_struct {
    vrpsx_system_t *system;
    vrpsx_mutex_t *mutex;
    vrpsx_cond_t *cond;
    int reference_count;
    sigwait_step_t sigwait_step;
    vrpsx_signo_t expected_signo;  // constant when shared across threads
    vrpsx_signo_t actual_signo;
} sigwait_state_t;

static int sigwait_state_delete(
    where_t where,
    sigwait_state_t *state)
{
    int result = 0;

    if (! state) {
        return 0;
    }

    if (state->cond && cond_disown(where, state->system, state->cond)) {
        result = -1;
    }

    if (state->mutex && mutex_disown(where, state->system, state->mutex)) {
        result = -1;
    }

    VRPSX_COUNTED_REFERENCE_CLEAR(vrpsx_system_t, &state->system);

    free(state);

    return result;
}

static int sigwait_state_create(
    where_t where,
    vrpsx_system_t *system,
    vrpsx_signo_t expected_signo,
    sigwait_state_t **statePtr)
{
    int result = -1;
    sigwait_state_t *state = NULL;

    *statePtr = NULL;

    state = MALLOC_AT(where, sigwait_state_t);
    if (! state) {
        goto fail;
    }
    memset(state, 0, sizeof *state);
    VRPSX_COUNTED_REFERENCE_ACQUIRE(vrpsx_system_t, &state->system, system);
    state->mutex = NULL;
    state->cond = NULL;
    state->reference_count = 1;
    state->sigwait_step = SIGWAIT_STEP_PREPARE;
    state->expected_signo = expected_signo;
    state->actual_signo = 0;

    if (mutex_create(where, system, &state->mutex)) {
        goto fail;
    }

    if (cond_create(where, system, &state->cond)) {
        goto fail;
    }

    result = 0;

cleanup:
    if (result) {
        sigwait_state_delete(where, state);
    } else {
        VRPSX_COUNTED_REFERENCE_MOVE(sigwait_state_t, statePtr, &state);
    }
    return result;

fail:
    result = -1;
    goto cleanup;
}

static void *sigwait_thread(
    void *arg)
{
    sigwait_state_t *state = (sigwait_state_t *)arg;
    vrpsx_system_t *system = state->system;
    vrpsx_signo_t actual_signo;
    int result = -1;
    char *space = NULL;
    int mutex_locked = 0;
    vrpsx_sigset_t *set;
    int reference_count;
    vrpsx_errno_t ern;

    space = (char *)malloc_at(HERE, VRPSX_CONSTANT(system, sizeof_sigset));
    if (! space) {
        goto fail;
    }

    if (sigset_init(HERE, system, space, &set)) {
        goto fail;
    }

    if (sigset_add(HERE, system, set, state->expected_signo)) {
        goto fail;
    }

    if (mutex_lock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 1;

    state->sigwait_step = SIGWAIT_STEP_IMMINENT;

    if (cond_broadcast(HERE, system, state->cond)) {
        goto fail;
    }

    if (mutex_unlock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 0;

    actual_signo = -1;
    ern = VRPSX_CALL2(system, sigwait, set, &actual_signo);
    if (ern) {
        printError(HERE, system, "sigwait", ern);
        goto fail;
    }

    if (mutex_lock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 1;

    state->actual_signo = actual_signo;
    state->sigwait_step = SIGWAIT_STEP_DONE;

    if (cond_broadcast(HERE, system, state->cond)) {
        goto fail;
    }

    if (mutex_unlock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 0;

    result = 0;

cleanup:
    if (! mutex_locked && mutex_lock(HERE, system, state->mutex)) {
        result = -1;
    }

    state->sigwait_step = SIGWAIT_STEP_DONE;

    reference_count = -- state->reference_count;

    if (cond_broadcast(HERE, system, state->cond)) {
        result = -1;
    }

    if (mutex_unlock(HERE, system, state->mutex)) {
        result = -1;
    }

    if (reference_count == 0 && sigwait_state_delete(HERE, state)) {
        result = -1;
    }

    free(space);
    return (void *)(intptr_t)result;

fail:
    result = -1;
    goto cleanup;
}

static int begin_sigwait(
    vrpsx_system_t *system,
    vrpsx_signo_t expected_signo,
    sigwait_state_t **statePtr,
    vrpsx_thread_t *tid)
{
    int result = -1;
    sigwait_state_t *state = NULL;
    int mutex_locked = 0;
    int kill_thread = 0;
    int reference_count;
    vrpsx_timespec_t duration;
    vrpsx_errno_t ern;

    *statePtr = NULL;

    if (sigwait_state_create(HERE, system, expected_signo, &state)) {
        goto fail;
    }

    ++ state->reference_count;
    ern = VRPSX_CALL3(system, thread_create, tid, &sigwait_thread, state);
    if (ern) {
        -- state->reference_count;
        printError(HERE, system, "thread_create", ern);
        goto fail;
    }
    kill_thread = 1;

    if (mutex_lock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 1;

    while (state->sigwait_step == SIGWAIT_STEP_PREPARE) {
        if (cond_wait(HERE, system, state->cond, state->mutex)) {
            goto fail;
        }
    }

    if (mutex_unlock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 0;

    /* Sleep a little to let the other thread actually enter "sigwait".
     * Sadly, we have no way to be sure when it actually gets in there.
     *
     * 0.1 sec, which we hope is long enough to exceed the time slice
     * granularity even on a virtual machine with coarse time slices,
     * for most test runs.
     */
    duration.tv_sec = 0;
    duration.tv_nsec = 100000000;

    ern = VRPSX_CALL2(system, nanosleep, &duration, NULL);
    if (ern) {
        printError(HERE, system, "nanosleep", ern);
        goto fail;
    }

    kill_thread = 0;
    result = 0;

cleanup:
    if (! mutex_locked && mutex_lock(HERE, system, state->mutex)) {
        result = -1;
    }

    if (result) {
        -- state->reference_count;
    }
    reference_count = state->reference_count;

    if (mutex_unlock(HERE, system, state->mutex)) {
        result = -1;
    }

    if (reference_count == 0 && sigwait_state_delete(HERE, state)) {
        result = -1;
    }

    if (kill_thread) {
        ern = VRPSX_CALL2(system, thread_kill, *tid, expected_signo);
        if (ern) {
            printError(HERE, system, "thread_kill", ern);
            result = -1;
        }

        ern = VRPSX_CALL1(system, thread_detach, *tid);
        if (ern) {
            printError(HERE, system, "thread_detach", ern);
            result = -1;
        }
    }

    if (! result) {
        VRPSX_COUNTED_REFERENCE_MOVE(sigwait_state_t, statePtr, &state);
    }
    return result;

fail:
    result = -1;
    goto cleanup;
}

static int end_sigwait(
    sigwait_state_t **statePtr,
    vrpsx_thread_t tid)
{
    int result = -1;
    sigwait_state_t *state;
    vrpsx_system_t *system;
    int mutex_locked = 0;
    int reference_count;
    vrpsx_signo_t actual_signo, expected_signo;
    vrpsx_errno_t ern;

    VRPSX_COUNTED_REFERENCE_MOVE(sigwait_state_t, &state, statePtr);
    system = state->system;

    if (mutex_lock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 1;

    while (state->sigwait_step != SIGWAIT_STEP_DONE) {
        if (cond_wait(HERE, system, state->cond, state->mutex)) {
            goto fail;
        }
    }

    actual_signo = state->actual_signo;
    expected_signo = state->expected_signo;

    if (mutex_unlock(HERE, system, state->mutex)) {
        goto fail;
    }
    mutex_locked = 0;

    if (actual_signo != expected_signo) {
        PRINT_HERE;
        printf("wrong signal number: %d (\"%s\") != %d (\"%s\")\n",
                actual_signo, strsignal(actual_signo),
                expected_signo, strsignal(expected_signo));
        goto fail;
    }

    result = 0;

cleanup:
    if (! mutex_locked && mutex_lock(HERE, system, state->mutex)) {
        result = -1;
    }

    reference_count = -- state->reference_count;

    if (mutex_unlock(HERE, system, state->mutex)) {
        result = -1;
    }

    if (reference_count == 0 && sigwait_state_delete(HERE, state)) {
        result = -1;
    }

    ern = VRPSX_CALL1(system, thread_detach, tid);
    if (ern) {
        printError(HERE, system, "thread_detach", ern);
        result = -1;
    }

    return result;

fail:
    result = -1;
    goto cleanup;
}

int test_sigwait(
    vrpsx_system_t *system)
{
    int result = -1;
    char *space = NULL;
    size_t sizeof_sigset = VRPSX_CONSTANT(system, sizeof_sigset);
    vrpsx_sigset_t *old_mask = NULL;
    vrpsx_sigset_t *to_block = NULL;
    int restore_mask = 0;
    sigwait_state_t *state = NULL;
    vrpsx_thread_t tid;
    vrpsx_errno_t ern;

    space = (char *)malloc_at(HERE, 2 * sizeof_sigset);
    if (! space) {
        goto fail;
    }

    if (sigset_init(HERE, system, space, &old_mask)) {
        goto fail;
    }

    if (sigset_init(HERE, system, space + sizeof_sigset, &to_block)) {
        goto fail;
    }

    if (sigset_add(HERE, system, to_block, VRPSX_CONSTANT(system, sigusr1))) {
        goto fail;
    }

    if (sigset_add(HERE, system, to_block, VRPSX_CONSTANT(system, sigusr2))) {
        goto fail;
    }

    ern = VRPSX_CALL3(system, thread_sigmask,
            VRPSX_CONSTANT(system, sig_block), to_block, old_mask);
    if (ern) {
        printError(HERE, system, "thread_sigmask",ern);
        goto fail;
    }
    restore_mask = 1;

    if (begin_sigwait(system, VRPSX_CONSTANT(system, sigusr2), &state, &tid)) {
        goto fail;
    }

    ern = VRPSX_CALL2(system, kill,
            VRPSX_CALL0(system, getpid),
            VRPSX_CONSTANT(system, sigusr2));
    if (ern) {
        printError(HERE, system, "kill", ern);
        goto fail;
    }

    if (end_sigwait(&state, tid)) {
        goto fail;
    }

    result = 0;

cleanup:
    if (state) {
        ern = VRPSX_CALL2(system, thread_kill, tid, state->expected_signo);
        if (ern) {
            printError(HERE, system, "thread_kill", ern);
        }

        end_sigwait(&state, tid);
    }

    if (restore_mask) {
        ern = VRPSX_CALL3(system, thread_sigmask,
                VRPSX_CONSTANT(system, sig_setmask), old_mask, NULL);
        if (ern) {
            printError(HERE, system, "thread_sigmask", ern);
        }
    }
    free(space);
    return result;

fail:
    result = -1;
    goto cleanup;
}


int test_thread_sigmask(
    vrpsx_system_t *system)
{
    int result = -1;
    char *space = NULL;
    size_t sizeof_sigset = VRPSX_CONSTANT(system, sizeof_sigset);
    vrpsx_sigset_t *old_mask = NULL;
    vrpsx_sigset_t *to_change = NULL;
    vrpsx_sigset_t *previous_mask = NULL;
    int restore_mask = 0;
    int actual_ismember;
    sigwait_state_t *state = NULL;
    vrpsx_thread_t tid;
    size_t const *signo_offset;
    vrpsx_errno_t ern;

    space = (char *)malloc_at(HERE, 3 * sizeof_sigset);
    if (! space) {
        goto fail;
    }

    if (sigset_init(HERE, system, space, &old_mask)) {
        goto fail;
    }

    if (sigset_init(HERE, system, space + sizeof_sigset, &to_change)) {
        goto fail;
    }

    if (sigset_init(HERE, system, space + 2 * sizeof_sigset, &previous_mask)) {
        goto fail;
    }

    if (sigset_add(HERE, system, to_change, VRPSX_CONSTANT(system, sigusr1))) {
        goto fail;
    }

    if (sigset_add(HERE, system, to_change, VRPSX_CONSTANT(system, sigusr2))) {
        goto fail;
    }

    ern = VRPSX_CALL3(system, thread_sigmask,
            VRPSX_CONSTANT(system, sig_block), to_change, old_mask);
    if (ern) {
        printError(HERE, system, "thread_sigmask", ern);
        goto fail;
    }
    restore_mask = 1;

    if (begin_sigwait(system, VRPSX_CONSTANT(system, sigusr1), &state, &tid)) {
        goto fail;
    }

    ern = VRPSX_CALL2(system, kill,
            VRPSX_CALL0(system, getpid),
            VRPSX_CONSTANT(system, sigusr1));
    if (ern) {
        printError(HERE, system, "kill", ern);
        goto fail;
    }

    if (end_sigwait(&state, tid)) {
        goto fail;
    }

    if (sigset_del(HERE, system, to_change, VRPSX_CONSTANT(system, sigusr1))) {
        goto fail;
    }

    ern = VRPSX_CALL3(system, thread_sigmask,
            VRPSX_CONSTANT(system, sig_unblock), to_change, previous_mask);
    if (ern) {
        printError(HERE, system, "thread_sigmask",ern);
        goto fail;
    }

    /* Check that we blocked sigusr1 and sigusr2 in the call
     * that used sig_block (NOT the immediately previous call).
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);

        result = sigset_ismember(
            HERE, system, old_mask, signo, &actual_ismember);
        if (result) {
            goto fail;
        }
        result = -1;

        if (signo == VRPSX_CONSTANT(system, sigusr1)
                || signo == VRPSX_CONSTANT(system, sigusr2)) {
            actual_ismember = 1;
        }

        if (assert_sigset_membership(
                    HERE, system, previous_mask, signo, actual_ismember)) {
            goto fail;
        }
    }

    ern = VRPSX_CALL3(system, thread_sigmask,
            VRPSX_CONSTANT(system, sig_setmask), old_mask, previous_mask);
    if (ern) {
        printError(HERE, system, "thread_sigmask",ern);
        goto fail;
    }

    /* Check that we unblocked sigusr2 in the call that used sig_unblock
     * (NOT the immediately previous call), but left sigusr1 blocked.
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);

        result = sigset_ismember(
            HERE, system, old_mask, signo, &actual_ismember);
        if (result) {
            goto fail;
        }
        result = -1;

        if (signo == VRPSX_CONSTANT(system, sigusr1)) {
            actual_ismember = 1;
        }

        if (signo == VRPSX_CONSTANT(system, sigusr2)) {
            actual_ismember = 0;
        }

        if (assert_sigset_membership(
                    HERE, system, previous_mask, signo, actual_ismember)) {
            goto fail;
        }
    }

    ern = VRPSX_CALL3(system, thread_sigmask,
            VRPSX_CONSTANT(system, sig_setmask), old_mask, previous_mask);
    if (ern) {
        printError(HERE, system, "thread_sigmask",ern);
        goto fail;
    }

    /* Check that we restored the old mask in the FIRST call
     * that used sig_setmask (NOT the immediately previous call).
     */
    for (signo_offset = signo_offsets; *signo_offset; ++signo_offset) {
        vrpsx_signo_t signo = signo_constant(system, *signo_offset);

        result = sigset_ismember(
            HERE, system, old_mask, signo, &actual_ismember);
        if (result) {
            goto fail;
        }
        result = -1;

        if (assert_sigset_membership(
                    HERE, system, previous_mask, signo, actual_ismember)) {
            goto fail;
        }
    }

    result = 0;

cleanup:
    if (state) {
        ern = VRPSX_CALL2(system, thread_kill, tid, state->expected_signo);
        if (ern) {
            printError(HERE, system, "thread_kill", ern);
        }

        end_sigwait(&state, tid);
    }

    if (restore_mask) {
        ern = VRPSX_CALL3(system, thread_sigmask,
                VRPSX_CONSTANT(system, sig_setmask), old_mask, NULL);
        if (ern) {
            printError(HERE, system, "thread_sigmask", ern);
        }
    }
    free(space);
    return result;

fail:
    result = -1;
    goto cleanup;
}


typedef int test_t(vrpsx_system_t *);

typedef struct test_definition_struct {
    char const *name;
    test_t *function;
} test_definition_t;

#define TEST_DEFINITION(NAME)                                   \
    { VRPSX_STRINGIZE(NAME), & VRPSX_CAT2(test_, NAME) }

static test_definition_t const allTests[] = {
    TEST_DEFINITION(error_acquire),
    TEST_DEFINITION(memory_allocation),
    TEST_DEFINITION(exit),
    TEST_DEFINITION(uname),
    TEST_DEFINITION(sysconf),
    TEST_DEFINITION(pathconf),
    TEST_DEFINITION(fpathconf),
    TEST_DEFINITION(getcwd),
    TEST_DEFINITION(get_timezone),
    TEST_DEFINITION(get_tzname),
    TEST_DEFINITION(gmtime),
    TEST_DEFINITION(localtime),
    TEST_DEFINITION(mktime),
    TEST_DEFINITION(time),
    TEST_DEFINITION(clock_gettime),
/* nanosleep_uninterrupted and nanosleep_interrupted are not reliable on GCP.
 * It's not even clear that we gain any significant advantage by running the
 * tests.  What are they testing?
 * 
 * TEST_DEFINITION(nanosleep_uninterrupted),
 * TEST_DEFINITION(nanosleep_interrupted),
 */
    TEST_DEFINITION(getlogin),
    TEST_DEFINITION(getpwnam),
    TEST_DEFINITION(getpid),
    TEST_DEFINITION(getpgrp),
    TEST_DEFINITION(getpgid),
    TEST_DEFINITION(kill),
    TEST_DEFINITION(raise),
    TEST_DEFINITION(strsignal),
    TEST_DEFINITION(sigset),
    TEST_DEFINITION(sigaction),
    TEST_DEFINITION(sigwait),
    TEST_DEFINITION(thread_sigmask),
    { NULL, NULL }
};


VRPSX_DECLARE_IDENTIFIED_FUNCTION(static, main, vrpsx_system_main);

static int VRPSX_IDENTIFIED_FUNCTION_NAME(main, vrpsx_system_main)(
    vrpsx_system_initial_t *initial)
{
    vrpsx_system_t *system = initial->system;
    test_definition_t const *test;
    int exit_code = EXIT_SUCCESS;
    test_timer_t timer;
    struct timespec t0, t1;

    test_timer_construct(&timer);

    for (test = allTests; test->name; ++test) {
        printf("%s:\n", test->name);

        if (clock_gettime(CLOCK_REALTIME, &t0)) {
            perror("clock_gettime");
            exit_code = 1;
            break;
        }

        if (test->function(system)) {
            exit_code = EXIT_FAILURE;
        }

        if (clock_gettime(CLOCK_REALTIME, &t1)) {
            perror("clock_gettime");
            exit_code = 1;
            break;
        }

        test_timer_add_difference(&timer, test->name, &t0, &t1);
    }

    test_timer_report(&timer);

    test_timer_destroy(&timer);

    return exit_code;
}

vrpsx_identified_function_t const VRPSX_IDENTIFIED_FUNCTION_LIST[] = {
    VRPSX_MENTION_IDENTIFIED_FUNCTION(main, vrpsx_system_main),
    VRPSX_IDENTIFIED_FUNCTION_LIST_TERMINATOR
};
