/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares utilities to compute the "recursive digest"
 * of a header file as defined by "vrpsx_identified.h".
 *
 * NOTE: The utilities declared by this header are meant to satisfy
 * the definition of "recursive digest" in "vrpsx_identified.h", NOT
 * to contribute to that definition.  If you see a discrepancy, then
 * probably you should fix these utilities, not "vrpsx_identified.h".
 *
 * Typically you use these utilities only when computing identifiers.
 * Most clients and implementations of interfaces do NOT link to the
 * utilities declared here.  What matters to clients and implementations
 * is the fact of which identifier was actually assigned to an interface,
 * not the manner in which it was assigned.
 *
 * For example, for compatibility reasons a client might use an identifier
 * that was assigned randomly, or by a buggy variant of this library, or by
 * an implementation of a different pseudorandom function.  The actual code
 * used to compute an identifier is irrelevant once that identifier is known.
 * Clients and implementations use that knowledge, not the computation itself.
 *
 * Thread Safety:
 *
 * In general these routines are NOT thread-safe,
 * though it may be possible to remedy that lack.
 */

#ifndef VRPSX_IDENTIFY_LIB_H
#define VRPSX_IDENTIFY_LIB_H 1

#include "vrpsx_ostreambuf.h"
#include "vrpsx_stdint.h"
#include <vrpsx_utility.h>
#include <sys/types.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* Hash entry for hash_map_t.  The key and value data come
 * just after the end of the declared portion of this struct.
 *
 * Interface Lifetime: single-owner.
 * Thread Safety: requires synchronization.
 */
typedef struct hash_entry_struct {
    struct hash_entry_vtbl_struct const *vtbl;

    /* Next entry in the same hash bucket.
     * This member should be used only by the hash map.
     */
    struct hash_entry_struct *next;

    /* Hash used to compute bucket index.
     *
     * This value is a function of the key.
     * Equal keys must have the same hash.
     */
    size_t hash;
} hash_entry_t;

typedef struct hash_entry_end_struct {
} hash_entry_end_t;

typedef struct hash_entry_vtbl_struct {
    /* Destroys and deallocates this interface.
     * As appropriate, deallocation occurs even if destruction fails.
     */
    void (*disown)(
        hash_entry_t *ent);

    /* Returns nonzero if and only if the key of this entry
     * is equal to the key of the other specified entry.
     *
     * WARNING: The entries must be implemented in
     * the same way or at least in compatible ways.
     */
    int (*equal)(
        hash_entry_t *ent,
        hash_entry_t *other);

    hash_entry_end_t const *hash_entry_end;
} hash_entry_vtbl_t;

/* A generic hash table.
 *
 * Thread Safety: requires synchronization.
 */
typedef struct hash_map_struct {
    /* Bucket array.
     */
    hash_entry_t **buckets;

    /* Number of buckets in bucket array.
     */
    size_t bucket_count;

    /* Number of entries: may be more or less than bucket count, though
     * the implementation tries to maintain the ratio in a certain range.
     */
    size_t entry_count;
} hash_map_t;

/* Initializes an empty map.
 */
void hash_map_init(
    hash_map_t *map);

/* Destroys a map, freeing all buckets and deleting all entries.
 */
void hash_map_destroy(
    hash_map_t *map);

/* Inserts an entry into a map unless a matching entry is already present.
 * Ownership if the new entry is transferred, and if a matching entry is
 * already present or a failure occurs then the given entry is disowned.
 *
 * Unless finalEnt == NULL, this function stores in *finalEnt a pointer to
 * the matching entry, whether new or old, WITHOUT transferring ownership.
 *
 * If insertion is required and memory allocation fails, then this function
 * stores NULL in *finalEnt (unless finalEnt == NULL) and returns a nonzero
 * error number.  Otherwise this function succeeds and returns 0.
 */
int hash_map_insert(
    hash_map_t *map,           /* Map into which to insert "ent". */
    hash_entry_t *newEnt,      /* The entry to insert; must be non-NULL. */
    int *created,              /* Set to 1 if new entry created;
                                  set to 0 if existing entry found
                                  or an error occurred.  Unused if NULL. */
    hash_entry_t **finalEnt);  /* After any insertion, this pointer is made
                                * to point to the matching entry.
                                * The map retains ownership of that entry.
                                * Set to NULL on memory allocation failure.
                                * Not used if NULL. */

/* Erases the given map entry, which must be in the given map.
 * Does not necessarily shrink the number of hash buckets.
 *
 * The given map must contain the given entry;
 * the program may crash if you violate this requirement.
 */
void hash_map_erase(
    hash_map_t *map,        /* Map containing given entry. */
    hash_entry_t *doomed);  /* Entry to erase from given map. */


/* Hash entry mapping a header file to a recursive digest,
 * or to all zeros while that digest is being computed.
 *
 * This type of entry can only be compared to entries
 * of the same type.
 *
 * Extends: hash_entry_t.
 * Interface Lifetime: single-owner.
 * Thread Safety: requires synchronization.
 */
typedef struct recursive_digest_entry_struct {
    hash_entry_t interface;

    /* The device containing the header file.
     * Do not modify this member after creation.
     */
    dev_t dev;

    /* The inode of the header file.
     * Do not modify this member after creation.
     */
    ino_t ino;

    /* The recursive digest if fully computed;
     * all zeros while computation is progressing.
     * You can modify this digest.
     */
    vrpsx_uint8_t digest[16];
} recursive_digest_entry_t;

/* Hash map of from headers visited by the recursive digest function to
 * their recursive digests, or to all zeros for visits still in progress.
 * The entries are all of type recursive_digest_entry_t.
 */
typedef struct recursive_digest_map_struct {
    hash_map_t generic;
} recursive_digest_map_t;

/* The recursive_digest_map_t specialization of hash_map_init.
 */
void recursive_digest_map_init(
    recursive_digest_map_t *map);

/* The recursive_digest_map_t specialization of hash_map_destroy.
 */
void recursive_digest_map_destroy(
    recursive_digest_map_t *map);

/* The recursive_digest_map_t specialization of hash_map_insert.
 * Handles allocation of the recursive_digest_entry_t entry.
 */
int recursive_digest_map_insert(
    recursive_digest_map_t *map,
    dev_t dev,                 /* Device containing header. */
    ino_t ino,                 /* Header file inode. */
    int *created,              /* Set to 1 if new entry created;
                                  set to 0 if existing entry found
                                  or an error occurred.  Unused if NULL. */
    recursive_digest_entry_t **finalEnt);

/* The recursive_digest_map_t specialization of hash_map_erase.
 */
void recursive_digest_map_erase(
    recursive_digest_map_t *map,
    recursive_digest_entry_t *doomed);


/* Computes the length of a header name such as you might find
 * in a C include directive, including opening and closing quotes.
 * For example, both <my.h> and "my.h" have length 6.
 */
size_t quoted_length_of_include(
    char const *qpath);  /* Relative or absolute header pathname, surrounded by
                          * double quotes or angle brackets to indicate which
                          * list of include directories to use if relative.
                          * Terminated by '"' or '>'.  No NUL is required. */

/* On success, stores in "digest" the "recursive digest" of the given
 * header file, as defined by "vrpsx_identified.h", and returns zero.
 *
 * On failure, the content of "digest" is all zeros and this function
 * returns a nonzero error number such as you might find in "errno".
 *
 * This function inserts the header into the map before making any
 * recursive calls, unless it is aleady in the map but not fully
 * hashed, in which this function fails with ELOOP.
 *
 * If error_message != NULL then this function sets *error_message to NULL
 * unless it successfully composed an error message for the failure, in which
 * case it stores that message, which the caller must deallocate with "free".
 * Composition fails only due to insufficient memory.
 */
int recursive_digest(
    char **error_message,
    recursive_digest_map_t *map,
    char const *const *include_directories, /* NULL-terminated list of dirs. */
    char const *context,                  /* If non-NULL, the including file.*/
    long context_line,                    /* If context != NULL, line number.*/
    char const *wd,                       /* Working dir for "..." includes. */
    char const *qpath,                    /* See quoted_length_of_include. */
    unsigned char digest[16],
    vrpsx_ostreambuf_t *debug);           /* If not NULL, copy digest input
                                           * to this debug output stream. */


/* On success, stores in "*interface_identifier" the interface identifier
 * of the given header file defined by the "Pseudorandom Identification"
 * suggestion in "vrpsx_identified.h", and returns zero.
 *
 * On failure, the "*interface_identifier" is 0 and this function
 * returns a nonzero error number such as you might find in "errno".
 * A header inclusion cycle that is not excluded results in error ELOOP.
 *
 * If error_message != NULL then this function sets *error_message to NULL
 * unless it successfully composed an error message for the failure, in which
 * case it stores that message, which the caller must deallocate with "free".
 * Composition fails only due to insufficient memory.
 */
int interface_identifier(
    char **error_message,
    char const *const *include_directories, /* NULL-terminated list of dirs. */
    char const *path,                      /* Pathname of definitive header. */
    char const *name,                      /* Interface name within header,
                                            * such as "vrpsx_identified". */
    vrpsx_uint32_t *interface_identifier, /* Must not be NULL. */
    vrpsx_ostreambuf_t *debug);           /* If not NULL, copy digest input
                                           * to this debug output stream. */


/* Presents an interface identifier as a string in the standard fashion.
 * See "vrpsx_identified.h" for details.
 */
void interface_identifier_to_string(
    vrpsx_uint32_t interface_identifier,
    char interface_identifier_string[9]);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_IDENTIFY_LIB_H */
