/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares the type to use for reference counts.
 */

#ifndef VRPSX_REFERENCE_COUNTED_H
#define VRPSX_REFERENCE_COUNTED_H 1

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* An interface whose lifetime is controlled by reference counting.
 *
 * Interface Lifetime: reference counted.
 * Thread Safety: thread-safe.
 */
typedef struct vrpsx_reference_counted_struct {
#define VRPSX_REFERENCE_COUNTED_MEMBERS(VTBL)   \
    VTBL const *vtbl;

    VRPSX_REFERENCE_COUNTED_MEMBERS(
        struct vrpsx_reference_counted_vtbl_struct)
} vrpsx_reference_counted_t;

typedef struct vrpsx_reference_counted_end_struct {
} vrpsx_reference_counted_end_t;

typedef struct vrpsx_reference_counted_vtbl_struct {
#define VRPSX_REFERENCE_COUNTED_VTBL_MEMBERS(TYPE)                      \
    /* NOTE: Unlike most interface operations, the first argument is    \
     * not a pointer to this interface; rather, it is a pointer to      \
     * such a pointer.  Both the argument and the interface pointer     \
     * must be non-NULL when you call this operation.                   \
     *                                                                  \
     * This operation sets (*counted) to NULL, atomically decrements    \
     * the reference count for this interface, and if it drops to 0     \
     * then releases any resources associated with this interface.      \
     *                                                                  \
     * If implementation resources are associated with more than one    \
     * interface, then the implementation is responsible for ensuring   \
     * their appropriately timed and reasonably prompt cleanup.         \
     *                                                                  \
     * WARNING: Implementations are free to maintain per-interface      \
     * reference counts if desired, and dispose of interfaces based     \
     * upon those counts.  Therefore an interface client must release   \
     * its reference to an interface THROUGH THAT SAME INTERFACE        \
     * (unless some interface extension clearly documents that the      \
     * implementation maintains just a one reference count for          \
     * all of the interfaces that it supports).                         \
     */                                                                 \
    void (*release)(                                                    \
        TYPE **counted);                                                \
                                                                        \
    /* Atomically increments the reference count for this interface.    \
     */                                                                 \
    void (*acquire)(                                                    \
        TYPE *counted);                                                 \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    vrpsx_reference_counted_end_t const *reference_counted_end;

    VRPSX_REFERENCE_COUNTED_VTBL_MEMBERS(vrpsx_reference_counted_t);
} vrpsx_reference_counted_vtbl_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_REFERENCE_COUNTED_H */
